from django.test.utils import override_settings

from germanium.decorators import data_consumer
from germanium.tools.trivials import assert_in, assert_equal, assert_true, assert_is_not_none
from germanium.tools.http import (assert_http_bad_request, assert_http_not_found, assert_http_method_not_allowed,
                                  assert_http_accepted, build_url)
from germanium.tools.rest import assert_valid_JSON_created_response, assert_valid_JSON_response

from .factories import UserFactory, IssueFactory
from .test_case import PystonTestCase

from app.elasticsearch.models import Comment


class ElasticsearchTestCase(PystonTestCase):

    COMMENT_API_URL = '/api/elasticsearch-comment/'

    @classmethod
    def setUpClass(cls):
        for i in range(10):
            comment = Comment(
                user_id=str(i),
                content=f'test message {i}',
                is_public=bool(i % 2),
                priority=i
            )
            comment.meta.id = i
            comment.save()
        Comment._index.refresh()

    @classmethod
    def tearDownClass(cls):
        Comment._index.delete()

    def test_get_elasticsearch_comments_should_return_right_data(self):
        resp = self.get(self.COMMENT_API_URL)
        assert_valid_JSON_response(resp)
        assert_equal(len(resp.json()), 10)
        for i, data in enumerate(resp.json()):
            assert_equal(data['priority'], i)
            assert_equal(data['user_id'], str(i))
            assert_equal(data['is_public'], bool(i % 2))
            assert_equal(data['content'], f'test message {i}')
            assert_equal(data['id'], str(i))

    def test_get_one_elasticsearch_comment_should_return_right_data(self):
        for i in range(10):
            resp = self.get(f'{self.COMMENT_API_URL}{i}/')
            assert_valid_JSON_response(resp)
            data = resp.json()
            assert_equal(data['priority'], i)
            assert_equal(data['user_id'], str(i))
            assert_equal(data['is_public'], bool(i % 2))
            assert_equal(data['content'], f'test message {i}')
            assert_is_not_none(data['id'])

    def test_get_elasticsearch_comments_should_be_sorted(self):
        resp = self.get(build_url(self.COMMENT_API_URL, order='priority'))
        assert_equal([v['priority'] for v in resp.json()], [i for i in range(10)])
        resp = self.get(build_url(self.COMMENT_API_URL, order='-priority'))
        assert_equal([v['priority'] for v in resp.json()], [i for i in range(10)][::-1])

        resp = self.get(build_url(self.COMMENT_API_URL, order='is_public'))
        assert_equal(resp.json()[0]['is_public'], False)
        resp = self.get(build_url(self.COMMENT_API_URL, order='-is_public'))
        assert_equal(resp.json()[0]['is_public'], True)

        resp = self.get(build_url(self.COMMENT_API_URL, order='user_id'))
        assert_equal([v['user_id'] for v in resp.json()], [str(i) for i in range(10)])
        resp = self.get(build_url(self.COMMENT_API_URL, order='-user_id'))
        assert_equal([v['user_id'] for v in resp.json()], [str(i) for i in range(10)][::-1])

        assert_http_bad_request(self.get(build_url(self.COMMENT_API_URL, order='content')))
        assert_http_bad_request(self.get(build_url(self.COMMENT_API_URL, order='id')))
        assert_http_bad_request(self.get(build_url(self.COMMENT_API_URL, order='invalid')))

    def test_get_elasticsearch_comments_should_be_filtered(self):
        resp = self.get(build_url(self.COMMENT_API_URL, filter='user_id=5'))
        assert_equal(len(resp.json()), 1)
        resp = self.get(build_url(self.COMMENT_API_URL, filter='content icontains "test message"'))
        assert_equal(len(resp.json()), 10)
        resp = self.get(build_url(self.COMMENT_API_URL, filter='content icontains "test message 1"'))
        assert_equal(len(resp.json()), 1)
        resp = self.get(build_url(self.COMMENT_API_URL, filter='is_public=1'))
        assert_equal(len(resp.json()), 5)
        assert_http_bad_request(self.get(build_url(self.COMMENT_API_URL.format(0), filter='invalid=5')))

    def test_get_elasticsearch_comments_should_paginated(self):
        headers = {'HTTP_X_OFFSET': '0', 'HTTP_X_BASE': '5'}
        resp = self.get(self.COMMENT_API_URL, headers=headers)
        assert_equal(len(self.deserialize(resp)), min(int(resp['x-total']), 5))

        headers = {'HTTP_X_OFFSET': '2', 'HTTP_X_BASE': '5'}
        resp = self.get(self.COMMENT_API_URL, headers=headers)
        assert_equal(len(self.deserialize(resp)), min(max(int(resp['x-total']) - 2, 0), 5))

        headers = {'HTTP_X_OFFSET': '2', 'HTTP_X_BASE': '-5'}
        resp = self.get(self.COMMENT_API_URL, headers=headers)
        assert_http_bad_request(resp)

        headers = {'HTTP_X_OFFSET': '-2', 'HTTP_X_BASE': '5'}
        resp = self.get(self.COMMENT_API_URL, headers=headers)
        assert_http_bad_request(resp)

        headers = {'HTTP_X_OFFSET': 'error', 'HTTP_X_BASE': 'error'}
        resp = self.get(self.COMMENT_API_URL, headers=headers)
        assert_http_bad_request(resp)



# Apply patch only if django is installed
try:
    from django.core.exceptions import ImproperlyConfigured
    try:
        from django.db import models  # noqa: F401
        from pyston import patch  # noqa: F401
    except ImproperlyConfigured:
        pass
except ImportError:
    pass



import types
import json

from io import StringIO

from collections import OrderedDict

from defusedxml import ElementTree as ET
from django.core.serializers.json import DjangoJSONEncoder
from django.http.response import HttpResponseBase
from django.template.loader import get_template
from django.utils.encoding import force_text
from django.utils.xmlutils import SimplerXMLGenerator
from django.utils.module_loading import import_string
from django.utils.html import format_html

from pyston.utils.helpers import UniversalBytesIO, serialized_data_to_python
from pyston.utils.datastructures import FieldsetGenerator
from pyston.conf import settings

from .file_generators import CsvGenerator, XlsxGenerator, PdfGenerator, TxtGenerator


def is_collection(data):
    return isinstance(data, (list, tuple, set, types.GeneratorType))


def get_default_converters():
    """
    Register all converters from settings configuration.
    """
    converters = OrderedDict()
    for converter_class_path in settings.CONVERTERS:
        converter_class = import_string(converter_class_path)()
        converters[converter_class.format] = converter_class
    return converters


def get_default_converter_name(converters=None):
    """
    Gets default converter name
    """
    converters = get_default_converters() if converters is None else converters
    return list(converters.keys())[0]


def get_converter(result_format, converters=None):
    """
    Gets an converter, returns the class and a content-type.
    """
    converters = get_default_converters() if converters is None else converters

    if result_format in converters:
        return converters.get(result_format)
    else:
        raise ValueError('No converter found for type {}'.format(result_format))


def get_converter_name_from_request(request, converters=None, input_serialization=False):
    """
    Function for determining which converter name to use
    for output.
    """
    try:
        import mimeparse
    except ImportError:
        mimeparse = None

    context_key = 'accept'
    if input_serialization:
        context_key = 'content_type'

    converters = get_default_converters() if converters is None else converters

    default_converter_name = get_default_converter_name(converters)

    if mimeparse and context_key in request._rest_context:
        supported_mime_types = set()
        converter_map = {}
        preferred_content_type = None
        for name, converter_class in converters.items():
            if name == default_converter_name:
                preferred_content_type = converter_class.media_type
            supported_mime_types.add(converter_class.media_type)
            converter_map[converter_class.media_type] = name
        supported_mime_types = list(supported_mime_types)
        if preferred_content_type:
            supported_mime_types.append(preferred_content_type)
        try:
            preferred_content_type = mimeparse.best_match(supported_mime_types,
                                                          request._rest_context[context_key])
        except ValueError:
            pass
        default_converter_name = converter_map.get(preferred_content_type, default_converter_name)

    return default_converter_name


def get_converter_from_request(request, converters=None, input_serialization=False):
    """
    Function for determining which converter name to use
    for output.
    """

    return get_converter(get_converter_name_from_request(request, converters, input_serialization), converters)


def get_supported_mime_types(converters):
    return [converter.media_type for _, converter in converters.items()]


class Converter:
    """
    Converter from standard data types to output format (JSON,YAML, Pickle) and from input to python objects
    """
    charset = 'utf-8'
    media_type = None
    format = None
    allow_tags = False

    @property
    def content_type(self):
        return '{}; charset={}'.format(self.media_type, self.charset)

    def _encode(self, data, options=None, **kwargs):
        """
        Encodes data to output string. You must implement this method or change implementation encode_to_stream method.
        """
        raise NotImplementedError

    def _decode(self, data, **kwargs):
        """
        Decodes data to string input
        """
        raise NotImplementedError

    def _encode_to_stream(self, output_stream, data, options=None, **kwargs):
        """
        Encodes data and writes it to the output stream
        """
        output_stream.write(self._encode(data, options=options, **kwargs))

    def encode_to_stream(self, output_stream, data, options=None, **kwargs):
        self._encode_to_stream(self._get_output_stream(output_stream), data, options=options, **kwargs)

    def decode(self, data, **kwargs):
        return self._decode(data, **kwargs)

    def _get_output_stream(self, output_stream):
        return output_stream if isinstance(output_stream, UniversalBytesIO) else UniversalBytesIO(output_stream)


class XmlConverter(Converter):
    """
    Converter for XML.
    Supports only output conversion
    """
    media_type = 'text/xml'
    format = 'xml'
    root_element_name = 'response'

    def _to_xml(self, xml, data):
        from pyston.serializer import LAZY_SERIALIZERS

        if isinstance(data, LAZY_SERIALIZERS):
            self._to_xml(xml, data.serialize())
        elif is_collection(data):
            for item in data:
                xml.startElement('resource', {})
                self._to_xml(xml, item)
                xml.endElement('resource')
        elif isinstance(data, dict):
            for key, value in data.items():
                xml.startElement(key, {})
                self._to_xml(xml, value)
                xml.endElement(key)
        else:
            xml.characters(force_text(data))

    def _encode(self, data, **kwargs):
        if data is not None:
            stream = StringIO()

            xml = SimplerXMLGenerator(stream, 'utf-8')
            xml.startDocument()
            xml.startElement(self.root_element_name, {})

            self._to_xml(xml, data)

            xml.endElement(self.root_element_name)
            xml.endDocument()

            return stream.getvalue()
        else:
            return ''

    def _decode(self, data, **kwargs):
        try:
            return ET.fromstring(data)
        except ET.ParseError as ex:
            raise ValueError(str(ex))


class LazyDjangoJsonEncoder(DjangoJSONEncoder):

    def default(self, o):
        from pyston.serializer import LAZY_SERIALIZERS

        if isinstance(o, types.GeneratorType):
            return tuple(o)
        elif isinstance(o, LAZY_SERIALIZERS):
            return o.serialize()
        else:
            return super(LazyDjangoJsonEncoder, self).default(o)


class JsonConverter(Converter):
    """
    JSON emitter, understands timestamps.
    """
    media_type = 'application/json'
    format = 'json'

    def _encode_to_stream(self, output_stream, data, options=None, **kwargs):
        options = settings.JSON_CONVERTER_OPTIONS if options is None else options
        if data is not None:
            json.dump(data, output_stream, cls=LazyDjangoJsonEncoder, ensure_ascii=False, **options)

    def _decode(self, data, **kwargs):
        return json.loads(data)


class GeneratorConverter(Converter):
    """
    Generator converter is more complicated.
    Contains user readable informations (headers).
    Supports only output.
    Output is flat.

    It is necessary set generator_class as class attribute

    This class contains little bit low-level implementation
    """

    generator_class = None

    def _render_headers(self, field_name_list):
        result = []
        if len(field_name_list) == 1 and '' in field_name_list:
            return result

        for field_name in field_name_list:
            result.append(field_name)
        return result

    def _get_recursive_value_from_row(self, data, key_path):
        from pyston.serializer import LAZY_SERIALIZERS

        if isinstance(data, LAZY_SERIALIZERS):
            return self._get_recursive_value_from_row(data.serialize(), key_path)
        elif len(key_path) == 0:
            return data
        elif isinstance(data, dict):
            return self._get_recursive_value_from_row(data.get(key_path[0], ''), key_path[1:])
        elif is_collection(data):
            return [self._get_recursive_value_from_row(val, key_path) for val in data]
        else:
            return ''

    def _render_dict(self, value, first):
        if first:
            return '\n'.join(('{}: {}'.format(key, self.render_value(val, False)) for key, val in value.items()))
        else:
            return '({})'.format(
                ', '.join(('{}: {}'.format(key, self.render_value(val, False)) for key, val in value.items()))
            )

    def _render_iterable(self, value, first):
        if first:
            return '\n'.join((self.render_value(val, False) for val in value))
        else:
            return '({})'.format(', '.join((self.render_value(val, False) for val in value)))

    def render_value(self, value, first=True):
        if isinstance(value, dict):
            return self._render_dict(value, first)
        elif is_collection(value):
            return self._render_iterable(value, first)
        else:
            return force_text(value)

    def _get_value_from_row(self, data, field):
        return self.render_value(self._get_recursive_value_from_row(data, field.key_path) or '')

    def _render_row(self, row, field_name_list):
        return (self._get_value_from_row(row, field) for field in field_name_list)

    def _render_content(self, field_name_list, converted_data):
        constructed_data = converted_data
        if not is_collection(constructed_data):
            constructed_data = [constructed_data]

        return (self._render_row(row, field_name_list) for row in constructed_data)

    def _encode_to_stream(self, output_stream, data, resource=None, requested_fields=None, direct_serialization=False,
                          **kwargs):
        fieldset = FieldsetGenerator(
            resource,
            force_text(requested_fields) if requested_fields is not None else None,
            direct_serialization=direct_serialization
        ).generate()
        self.generator_class().generate(
            self._render_headers(fieldset),
            self._render_content(fieldset, data),
            output_stream
        )


class CsvConverter(GeneratorConverter):
    """
    Converter for CSV response.
    Supports only output conversion
    """

    generator_class = CsvGenerator
    media_type = 'text/csv'
    format = 'csv'
    allow_tags = True


class XlsxConverter(GeneratorConverter):
    """
    Converter for XLSX response.
    For its use must be installed library xlsxwriter
    Supports only output conversion
    """

    generator_class = XlsxGenerator
    media_type = 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet'
    format = 'xlsx'
    allow_tags = True


class PdfConverter(GeneratorConverter):
    """
    Converter for PDF response.
    For its use must be installed library pisa
    Supports only output conversion
    """

    generator_class = PdfGenerator
    media_type = 'application/pdf'
    format = 'pdf'


class TxtConverter(GeneratorConverter):
    """
    Converter for TXT response.
    Supports only output conversion
    """

    generator_class = TxtGenerator
    media_type = 'plain/text'
    format = 'txt'
    allow_tags = True


class HtmlConverter(Converter):
    """
    Converter for HTML.
    Supports only output conversion and should be used only for debug
    """

    media_type = 'text/html'
    format = 'html'
    template_name = 'pyston/html_converter.html'

    def _get_put_form(self, resource, obj):
        from pyston.resource import BaseModelResource

        return (
            resource._get_form(inst=obj)
            if isinstance(resource, BaseModelResource) and resource.has_put_permission(obj=obj)
            else None
        )

    def _get_post_form(self, resource, obj):
        from pyston.resource import BaseModelResource

        return (
            resource._get_form(inst=obj)
            if isinstance(resource, BaseModelResource) and resource.has_post_permission(obj=obj)
            else None
        )

    def _get_forms(self, resource, obj):
        return {
            'post': self._get_post_form(resource, obj),
            'put': self._get_put_form(resource, obj),
        }

    def _get_converter(self, resource):
        return JsonConverter()

    def _get_permissions(self, resource, obj):
        return {
            'post': resource.has_post_permission(obj=obj),
            'get': resource.has_get_permission(obj=obj),
            'put': resource.has_put_permission(obj=obj),
            'delete': resource.has_delete_permission(obj=obj),
            'head': resource.has_head_permission(obj=obj),
            'options': resource.has_options_permission(obj=obj),
        } if resource else {}

    def _update_headers(self, http_headers, resource, converter):
        http_headers['Content-Type'] = converter.content_type
        return http_headers

    def encode_to_stream(self, output_stream, data, options=None, **kwargs):
        assert output_stream is not HttpResponseBase, 'Output stream must be http response'

        self._get_output_stream(output_stream).write(
            self._encode(data, response=output_stream, options=options, **kwargs)
        )

    def _convert_url_to_links(self, data):
        if isinstance(data, list):
            return [self._convert_url_to_links(val) for val in data]
        elif isinstance(data, dict):
            return OrderedDict((
                (key, format_html('<a href=\'{0}\'>{0}</a>', val) if key == 'url' else self._convert_url_to_links(val))
                for key, val in data.items()
            ))
        else:
            return data

    def _encode(self, data, response=None, http_headers=None, resource=None, result=None, **kwargs):
        from pyston.resource import BaseModelResource

        http_headers = {} if http_headers is None else http_headers.copy()
        converter = self._get_converter(resource)

        http_headers = self._update_headers(http_headers, resource, converter)
        obj = (
            resource._get_obj_or_none() if isinstance(resource, BaseModelResource) and resource.has_permission()
            else None
        )

        kwargs.update({
            'http_headers': http_headers,
            'resource': resource,
        })

        data_stream = UniversalBytesIO()
        converter._encode_to_stream(data_stream, self._convert_url_to_links(serialized_data_to_python(data)), **kwargs)

        context = kwargs.copy()
        context.update({
            'permissions': self._get_permissions(resource, obj),
            'forms': self._get_forms(resource, obj),
            'output': data_stream.getvalue(),
            'name': resource._get_name() if resource and resource.has_permission() else response.status_code
        })

        # All responses has set 200 response code, because response can return status code without content (204) and
        # browser doesn't render it
        response.status_code = 200

        return get_template(self.template_name).render(context, request=resource.request if resource else None)



from pyston.utils import LOOKUP_SEP
from pyston.filters.filters import Filter
from pyston.filters.exceptions import OperatorFilterError
from pyston.filters.managers import BaseParserModelFilterManager
from pyston.filters.filters import OPERATORS


class BaseDynamoFilter(Filter):

    allowed_operators = None

    def get_allowed_operators(self):
        return self.allowed_operators

    def clean_value(self, value, operator_slug, request):
        return value

    def get_q(self, value, operator_slug, request):
        if operator_slug not in self.get_allowed_operators():
            raise OperatorFilterError
        else:
            return self.get_filter_term(self.clean_value(value, operator_slug, request), operator_slug, request)

    def get_filter_term(self, value, operator_slug, request):
        raise NotImplementedError


class DynamoFilterManager(BaseParserModelFilterManager):

    def _logical_conditions_and(self, condition_a, condition_b):
        conditions_union = {**condition_a, **condition_b}
        sorted_keys = sorted(conditions_union)
        if len(condition_a) == 1 and len(condition_b) == 1 and len(sorted_keys) == 2:
            condition_a_full_identifier, condition_b_full_identifier = sorted_keys
            condition_a_identifier, condition_a_operator = condition_a_full_identifier.rsplit(LOOKUP_SEP, 1)
            condition_b_identifier, condition_b_operator = condition_b_full_identifier.rsplit(LOOKUP_SEP, 1)
            if (condition_a_identifier == condition_b_identifier and condition_a_operator == OPERATORS.GTE
                    and condition_b_operator == OPERATORS.LT):
                return {
                    f'{condition_a_identifier}__between': (
                        conditions_union[condition_a_full_identifier], conditions_union[condition_b_full_identifier]
                    )
                }
        return super()._logical_conditions_and(condition_a, condition_b)

    def _filter_queryset(self, qs, q):
        return qs.filter(**q)



import os

import csv
import codecs

from datetime import datetime, date
from decimal import Decimal

from django.conf import settings as django_settings
from django.utils.encoding import force_text
from django.template.loader import get_template

try:
    # xlsxwriter isn't standard with python.  It shouldn't be required if it
    # isn't used.
    import xlsxwriter
except ImportError:
    xlsxwriter = None
    XlsxGenerator = None


try:
    # pisa isn't standard with python. It shouldn't be required if it isn't used.
    from xhtml2pdf import pisa
except ImportError:
    pisa = None
    PdfGenerator = None

from pyston.conf import settings


TWOPLACES = Decimal(10) ** -2


class CsvGenerator:

    def __init__(self, delimiter=chr(59), quotechar=chr(34), quoting=csv.QUOTE_ALL, encoding='utf-8', **kwargs):
        self.encoding = encoding
        self.quotechar = quotechar
        self.quoting = quoting
        self.delimiter = delimiter
        self.kwargs = kwargs

    def generate(self, header, data, output_stream):
        writer = StreamCSV(
            output_stream,
            delimiter=self.delimiter,
            quotechar=self.quotechar,
            quoting=self.quoting,
            **self.kwargs
        )

        if header:
            writer.writerow(self._prepare_list(header))

        for row in data:
            writer.writerow(self._prepare_list(row))

    def _prepare_list(self, values):
        prepared_row = []

        for value in values:
            value = self._prepare_value(value.get('value') if isinstance(value, dict) else value)
            prepared_row.append(value)
        return prepared_row

    def _prepare_value(self, value):
        if isinstance(value, float):
            value = ('%.2f' % value).replace('.', ',')
        elif isinstance(value, Decimal):
            value = force_text(value.quantize(TWOPLACES)).replace('.', ',')
        else:
            value = force_text(value)
        return value.replace('&nbsp;', ' ')


class StreamCSV:

    def __init__(self, f, dialect=csv.excel, use_bom=True, **kwargs):
        self.writer = csv.writer(f, dialect=dialect, **kwargs)
        self.stream = f
        if use_bom:
            self.stream.write(force_text(codecs.BOM_UTF8))  # BOM for Excel

    def writerow(self, row):
        self.writer.writerow(row)
        self.stream.flush()

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)


class TxtGenerator:

    def _prepare_value(self, value):
        return value.replace('&nbsp;', ' ')

    def generate(self, header, data, output_stream):
        output_stream.write('---\n')
        for data_row in data:
            output_stream.write('\n')
            for col, val in enumerate(data_row):
                if header:
                    output_stream.write('{}:\n'.format(header[col]))
                if isinstance(val, str):
                    val = self._prepare_value(val)
                output_stream.write('\t'.join(('\t' + force_text(val).lstrip()).splitlines(True)) + '\n\n')
            output_stream.write('---\n')


if xlsxwriter:
    class XlsxGenerator:

        def _prepare_value(self, value):
            return value.replace('&nbsp;', ' ')

        def generate(self, header, data, output_stream):
            wb = xlsxwriter.Workbook(output_stream, {'strings_to_formulas': False, 'strings_to_urls': False})
            ws = wb.add_worksheet()

            date_format = wb.add_format({'num_format': 'd. mmmm yyyy'})
            datetime_format = wb.add_format({'num_format': 'd. mmmm yyyy hh:mm:ss'})
            decimal_format = wb.add_format({'num_format': '0.00'})

            row = 0

            if header:
                for col, head in enumerate(header):
                    ws.write(row, col, force_text(head))
                row += 1

            for data_row in data:
                for col, val in enumerate(data_row):
                    if isinstance(val, datetime):
                        ws.write(row, col, val.replace(tzinfo=None), datetime_format)
                    elif isinstance(val, date):
                        ws.write(row, col, val, date_format)
                    elif isinstance(val, (Decimal, float)):
                        ws.write(row, col, val, decimal_format)
                    elif isinstance(val, str):
                        val = self._prepare_value(val)
                        ws.write(row, col, val)
                    else:
                        ws.write(row, col, val)
                row += 1
            wb.close()

if pisa:
    class PdfGenerator:

        encoding = 'utf-8'

        def generate(self, header, data, output_stream):
            def fetch_resources(uri, rel):
                urls = {
                    django_settings.MEDIA_ROOT: settings.MEDIA_URL,
                    django_settings.STATICFILES_ROOT: django_settings.STATIC_URL
                }
                for k, v in urls.items():
                    if (uri.startswith(v)):
                        return os.path.join(k, uri.replace(v, ""))
                return ''

            pisa.pisaDocument(
                force_text(
                    get_template(settings.PDF_EXPORT_TEMPLATE).render(
                        {'pagesize': 'A4', 'headers': header, 'data': data}
                    )
                ),
                output_stream, encoding=self.encoding, link_callback=fetch_resources
            )



from decimal import Decimal, InvalidOperation

from django.core.validators import validate_ipv4_address, validate_ipv46_address
from django.core.exceptions import ValidationError
from django.utils.translation import ugettext
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import make_aware

from chamber.utils.datastructures import Enum

from dateutil.parser import DEFAULTPARSER

from pyston.utils import LOOKUP_SEP

from .exceptions import FilterValueError, OperatorFilterError


OPERATORS = Enum(
    ('GT', 'gt'),
    ('LT', 'lt'),
    ('EQ', 'eq'),
    ('NEQ', 'neq'),
    ('LTE', 'lte'),
    ('GTE', 'gte'),
    ('CONTAINS', 'contains'),
    ('ICONTAINS', 'icontains'),
    ('RANGE', 'range'),
    ('EXACT', 'exact'),
    ('IEXACT', 'iexact'),
    ('STARTSWITH', 'startswith'),
    ('ISTARTSWITH', 'istartswith'),
    ('ENDSWITH', 'endswith'),
    ('IENDSWITH', 'iendswith'),
    ('IN', 'in'),
    ('RANGE', 'range'),
    ('ALL', 'all'),
    ('ISNULL', 'isnull'),
)


NONE_LABEL = _('(None)')


class Operator:
    """
    Operator is used for specific type of filters that allows more different ways how to filter queryset data according
    to input operator between identifier and value.
    """

    def get_q(self, value, request):
        """
        Method must be implemented inside descendant and should return django db Q object that will be used for purpose
        of filtering queryset.
        """
        raise NotImplementedError


class Filter:
    """
    Filters purpose is return Q object that is used for filtering data that resource returns.
    Filter can be joined to the field, method or resource.
    """

    suffixes = {}
    choices = None
    allowed_operators = {}

    def __init__(self, identifiers_prefix, identifiers, identifiers_suffix, model, field=None, method=None):
        """
        Filter init values are these:
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param identifiers_suffix: list of suffixes that can be used for more specific filters.
               For example for a date filter can be used suffixes month, day, year.
        :param model: Django model class of filtered object.
        :param field: model field which is related with filter.
        :param method: method that is related with filter.
        method and field cannot be set together.
        """
        assert field is None or method is None, 'Field and method cannot be set together'

        self.identifiers_prefix = identifiers_prefix
        self.identifiers = identifiers
        self.identifiers_suffix = identifiers_suffix
        self.full_identifiers = identifiers_prefix + identifiers + identifiers_suffix
        self.field = field
        self.method = method
        self.model = model

    def clean_value(self, value, operator_slug, request):
        """
        Method that cleans input value to the filter specific format.
        """
        return value

    def get_full_filter_key(self):
        return LOOKUP_SEP.join(self.full_identifiers)

    def get_allowed_operators(self):
        """
        :return: list of allowed operators to the concrete filter.
        """
        return self.allowed_operators

    @classmethod
    def get_suffixes(cls):
        """
        :return: set of allowed suffixes for the operator.
        """
        return cls.suffixes

    def get_q(self, value, operator_slug, request):
        """
        Method must be implemented inside descendant and should return django db Q object that will be used for purpose
        of filtering rest response data.
        """
        raise NotImplementedError


class OperatorsFilterMixin:
    """
    Mixin is used for specific type of filter that uses operator objects.
    """

    operators = ()

    def get_allowed_operators(self):
        return [operator_key for operator_key, operator in self.operators]

    def get_operator_obj(self, operator_slug):
        """
        :return: concrete operator object for the specific operator key.
        """
        operator_obj = dict(self.operators).get(operator_slug)
        if not operator_obj:
            raise OperatorFilterError
        return operator_obj

    def get_q(self, value, operator_slug, request):
        operator_obj = self.get_operator_obj(operator_slug)
        return operator_obj.get_q(self, value, operator_slug, request)


class MethodFilter(Filter):
    """
    Abstract parent for all method filters.
    """

    def __init__(self, identifiers_prefix, identifiers, identifiers_suffix, model, method=None):
        assert method, 'Method is required'
        super().__init__(identifiers_prefix, identifiers, identifiers_suffix, model, method=method)


class ModelFieldFilter(Filter):
    """
    Abstract parent for all model field filters.
    """

    def __init__(self, identifiers_prefix, identifiers, identifiers_suffix, model, field=None):
        assert field, 'Field is required'
        super().__init__(identifiers_prefix, identifiers, identifiers_suffix, model, field=field)


class OperatorsModelFieldFilter(OperatorsFilterMixin, ModelFieldFilter):
    pass


class BooleanFilterMixin:
    """
    Helper that contains cleaner for boolean input values.
    """

    choices = (
        (1, _('Yes')),
        (0, _('No'))
    )

    def clean_value(self, value, operator_slug, request):
        if isinstance(value, bool):
            return value
        elif value in {'true', 'false'}:
            return value == 'true'
        elif value in {'0', '1'}:
            return value == '1'
        else:
            raise FilterValueError(ugettext('Value must be boolean'))


class NullBooleanFilterMixin(BooleanFilterMixin):
    """
    Helper that contains cleaner for boolean input values where value can be None.
    """

    choices = (
        (None, NONE_LABEL),
        (1, _('Yes')),
        (0, _('No'))
    )

    def clean_value(self, value, operator_slug, request):
        if value is None:
            return value
        else:
            return super().clean_value(value, operator_slug, request)


class IntegerFilterMixin:
    """
    Helper that contains cleaner for integer input values.
    """

    def clean_value(self, value, operator_slug, request):
        if value is None:
            return value
        try:
            return int(value)
        except (ValueError, TypeError):
            raise FilterValueError(ugettext('Value must be integer'))


class FloatFilterMixin:

    def clean_value(self, value, operator_slug, request):
        if value is None:
            return value
        try:
            return float(value)
        except (ValueError, TypeError):
            raise FilterValueError(ugettext('Value must be float'))


class DecimalFilterMixin:

    def clean_value(self, value, operator_slug, request):
        if value is None:
            return value
        try:
            return Decimal(value)
        except InvalidOperation:
            raise FilterValueError(ugettext('Value must be decimal'))


class IPAddressFilterMixin:

    def clean_value(self, value, operator_slug, request):
        if value is None:
            return value
        elif operator_slug not in {OPERATORS.CONTAINS, OPERATORS.EXACT, OPERATORS.STARTSWITH, OPERATORS.ENDSWITH}:
            try:
                validate_ipv4_address(value)
            except ValidationError:
                raise FilterValueError(ugettext('Value must be in format IPv4.'))
        return value


class GenericIPAddressFilterMixin:

    def clean_value(self, value, operator_slug, request):
        if value is None:
            return value
        elif operator_slug not in {OPERATORS.CONTAINS, OPERATORS.EXACT, OPERATORS.STARTSWITH, OPERATORS.ENDSWITH}:
            try:
                validate_ipv46_address(value)
            except ValidationError:
                raise FilterValueError(ugettext('Value must be in format IPv4 or IPv6.'))
        return value


class DateFilterMixin:

    suffixes = {
        'day', 'month', 'year'
    }

    def _clean_datetime_to_parts(self, value):
        value = DEFAULTPARSER._parse(value, dayfirst='-' not in value)
        value = value[0] if isinstance(value, tuple) else value

        if value is None:
            raise FilterValueError(ugettext('Value cannot be parsed to partial datetime'))
        else:
            return value

    def _clean_integer(self, value):
        try:
            return int(value)
        except ValueError:
            raise FilterValueError(ugettext('Value must be integer'))

    def _clean_datetime(self, value):
        try:
            datetime_value = DEFAULTPARSER.parse(value, dayfirst='-' not in value)
            return make_aware(datetime_value, is_dst=True) if datetime_value.tzinfo is None else datetime_value
        except ValueError:
            raise FilterValueError(ugettext('Value must be in format ISO 8601.'))

    def clean_value(self, value, operator_slug, request):
        suffix = self.identifiers_suffix[0] if self.identifiers_suffix else None
        if suffix in self.suffixes:
            return self._clean_integer(value)
        elif operator_slug == OPERATORS.CONTAINS:
            return self._clean_datetime_to_parts(value)
        elif value is None:
            return value
        else:
            return self._clean_datetime(value)



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2016-12-26 15:01
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Issue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('name', models.CharField(max_length=100, verbose_name='name')),
                ('description', models.TextField(blank=True, null=True, verbose_name='description')),
            ],
        ),
        migrations.CreateModel(
            name='User',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='created at')),
                ('email', models.EmailField(max_length=254, unique=True, verbose_name='email')),
                ('contract', models.FileField(blank=True, null=True, upload_to='documents/', verbose_name='file')),
                ('is_superuser', models.BooleanField(default=True, verbose_name='is superuser')),
                ('first_name', models.CharField(blank=True, max_length=100, null=True, verbose_name='first name')),
                ('last_name', models.CharField(blank=True, max_length=100, null=True, verbose_name='last name')),
            ],
        ),
        migrations.AddField(
            model_name='issue',
            name='created_by',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='created_issues', to='app.User', verbose_name='created by'),
        ),
        migrations.AddField(
            model_name='issue',
            name='leader',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='leading_issue', to='app.User', verbose_name='leader'),
        ),
        migrations.AddField(
            model_name='issue',
            name='solver',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='solving_issue', to='app.User', verbose_name='solver'),
        ),
        migrations.AddField(
            model_name='issue',
            name='watched_by',
            field=models.ManyToManyField(blank=True, related_name='watched_issues', to='app.User', verbose_name='watched by'),
        ),
    ]



from dj.settings.base import *

DEBUG = TEMPLATES[0]['OPTIONS']['debug'] = THUMBNAIL_DEBUG = True

ALLOWED_HOSTS = ['localhost']

# URL with protocol (and port)
PROJECT_URL = 'localhost:8000'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(PROJECT_DIR, 'var', 'db', 'sqlite.db'),
        'USER': '',
        'PASSWORD': '',
    },
}

ELASTICSEARCH_DATABASE = dict(
    hosts=[{'host': 'localhost', 'port': 9200}],
)

PYDJAMODB_DATABASE = {
    'HOST': 'http://localhost:8000',
    'AWS_ACCESS_KEY_ID': '_',
    'AWS_SECRET_ACCESS_KEY': '_',
    'AWS_REGION': None,
    'TABLE_PREFIX': 'pyston',
    'BILLING_MODE': 'PAY_PER_REQUEST',
}



from pyston.utils import RFS


class RequestedFieldsParser:
    """
    Abstract fields parser.
    """

    def parse(self, request):
        """
        :param request: Django HTTP request.
        :return: returns RFS objects..
        """
        raise NotImplementedError


class DefaultRequestedFieldsParser:
    """
    Default fields parserr.
    E.q.:
        /api/user?_filter=first_name,last_name,issues(id)
    """

    def parse(self, request):
        input = request._rest_context.get('fields')
        if input is None:
            return None
        return RFS.create_from_string(input)



from unittest.case import TestCase

from germanium.tools import assert_true, assert_false, assert_equal, assert_is_none

from pyston.utils import rfs


class FieldsetsTestCase(TestCase):

    def test_create_rfs_from_list(self):
        assert_equal(str(rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',))))), 'a,b(c,g),d(e(f))')
        assert_equal(str(rfs(('a', 'b', ('d', ('e__f',)), ('b', ('c',))))), 'a,b(c),d(e(f))')
        assert_equal(str(rfs(())), '')

    def test_rfs_append(self):
        fieldset = rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',))))
        fieldset.append('a')
        assert_equal(str(fieldset), 'a,b(c,g),d(e(f))')
        fieldset.append('a__h')
        assert_equal(str(fieldset), 'a(h),b(c,g),d(e(f))')
        fieldset.append('b__c__i')
        assert_equal(str(fieldset), 'a(h),b(c(i),g),d(e(f))')
        fieldset.append(('b', ('k')))
        assert_equal(str(fieldset), 'a(h),b(c(i),g,k),d(e(f))')

    def test_rfs_update(self):
        fieldset_a = rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',))))
        fieldset_b = rfs(('a__i', 'l'))

        fieldset_a.update(fieldset_b)
        assert_equal(str(fieldset_a), 'a(i),b(c,g),d(e(f)),l')
        assert_equal(str(fieldset_b), 'a(i),l')

    def test_rfs_update_with_list(self):
        fieldset_a = rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',))))
        fieldset_b = ('a__i', 'l')

        fieldset_a.update(fieldset_b)
        assert_equal(str(fieldset_a), 'a(i),b(c,g),d(e(f)),l')

    def test_rfs_flat(self):
        assert_equal(rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',)))).flat(), {'a', 'b', 'd'})

    def test_rfs_bool(self):
        assert_true(rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',)))))
        assert_false(rfs())
        assert_false(rfs(()))
        assert_false(rfs({}))

    def test_rfs_add(self):
        fieldset_a = rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',))))
        fieldset_b = rfs(('a__i', 'l'))

        fieldset_c = fieldset_a + fieldset_b
        assert_equal(str(fieldset_a), 'a,b(c,g),d(e(f))')
        assert_equal(str(fieldset_b), 'a(i),l')
        assert_equal(str(fieldset_c), 'a(i),b(c,g),d(e(f)),l')

    def test_rfs_add_list(self):
        fieldset_a = rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',))))
        fieldset_b = ('a__i', 'l')

        fieldset_c = fieldset_a + fieldset_b
        assert_equal(str(fieldset_a), 'a,b(c,g),d(e(f))')
        assert_equal(str(fieldset_c), 'a(i),b(c,g),d(e(f)),l')

    def test_rfs_get(self):
        fieldset = rfs(('a', 'b', 'b__c', 'b__g', ('d', ('e__f',))))
        assert_equal(str(fieldset.get('b')), 'b(c,g)')
        assert_is_none(fieldset.get('k'))
        assert_equal(str(fieldset.get('a')), 'a')

    def test_rfs_intersection(self):
        fieldset_a = rfs((('a', ('i', 'j')), 'b', 'b__c', 'b__g', ('d', ('e__f',))))
        fieldset_b = rfs(('a__i', 'b', 'l'))

        assert_equal(str(fieldset_a.intersection(fieldset_b)), 'a(i),b')



from germanium.tools.trivials import assert_equal
from germanium.tools.http import assert_http_bad_request, build_url
from germanium.tools.rest import assert_valid_JSON_response

from .factories import IssueFactory, UserFactory
from .test_case import PystonTestCase


class OrderTestCase(PystonTestCase):

    def test_order_by_decorator(self):
        [IssueFactory(description=str(i)) for i in range(10)]
        data = self.deserialize(self.get(build_url(self.ISSUE_API_URL, order='short_description')))
        assert_equal([v['short_description'] for v in data], [str(i) for i in range(10)])
        data = self.deserialize(self.get(build_url(self.ISSUE_API_URL, order='-short_description')))
        assert_equal([v['short_description'] for v in data], [str(i) for i in range(10)][::-1])
        assert_valid_JSON_response(self.get(build_url(self.USER_API_URL, order='solving_issue__short_description')))
        assert_valid_JSON_response(self.get(build_url(self.USER_API_URL, order='-solving_issue__short_description')))
        assert_http_bad_request(self.get(build_url(self.ISSUE_API_URL, order='description')))

    def test_override_extra_order_fields(self):
        assert_http_bad_request(self.get(build_url(self.USER_API_URL, order='created_at')))
        assert_valid_JSON_response(
            self.get(build_url(self.USER_API_URL, order='email,-solving_issue__short_description'))
        )

    def test_issue_can_order_only_with_readable_fields_and_extra_field(self):
        assert_valid_JSON_response(self.get(build_url(self.ISSUE_API_URL, order='solver__created_at')))
        assert_valid_JSON_response(self.get(build_url(self.ISSUE_API_URL, order='created_at')))
        assert_http_bad_request(self.get(build_url(self.ISSUE_API_URL, order='solver__manual_created_date')))

    def test_extra_sorter(self):
        user1 = UserFactory()
        user2 = UserFactory()
        user3 = UserFactory()
        user1.watched_issues.add(*(IssueFactory() for _ in range(2)))
        user3.watched_issues.add(*(IssueFactory() for _ in range(5)))

        users_pks = {user1.pk, user2.pk, user3.pk}

        resp = self.get(build_url(self.USER_API_URL, order='watched_issues_count'))
        assert_valid_JSON_response(resp)
        assert_equal(self.get_pk_list(resp, only_pks=users_pks), [user2.pk, user1.pk, user3.pk])

        resp = self.get(build_url(self.USER_API_URL, order='-watched_issues_count'))
        assert_valid_JSON_response(resp)
        assert_equal(self.get_pk_list(resp, only_pks=users_pks), [user3.pk, user1.pk, user2.pk])



from django.db.models import Q
from django.core.exceptions import FieldDoesNotExist
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext

from pyston.exception import RestException
from pyston.utils import rfs, LOOKUP_SEP
from pyston.utils.helpers import get_field_or_none, get_method_or_none
from pyston.serializer import get_resource_or_none

from .exceptions import FilterValueError, OperatorFilterError, FilterIdentifierError
from .utils import LOGICAL_OPERATORS
from .parser import QueryStringFilterParser, DefaultFilterParser, FilterParserError
from .django_filters import get_default_field_filter_class


def get_allowed_filter_fields_rfs_from_model(model):
    return rfs(model._rest_meta.extra_filter_fields).join(rfs(model._rest_meta.filter_fields))


class BaseModelFilterManager:
    """
    Filter manager is used inside object resource for composing filters with purpose to restrict output data according
    to input values.
    This is abstract class that provides methods to obtain concrete filters from resource and model methods and fields.
    """

    def _get_method_filter(self, method, identifiers_prefix, identifiers, identifiers_suffix, model, resource, request,
                           filters_fields_rfs):
        """
        :param method: method from which we can get filter.
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param identifiers_suffix: list of suffixes that can be used for more specific filters.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param filters_fields_rfs: RFS of fields that is allowed to filter.
        :return: method returns filter object that is obtained from method.
        """
        if hasattr(method, 'filter_by'):
            # If method has filter_by attribute filter is being searched according to this value.
            filter_by_identifiers = method.filter_by.split(LOOKUP_SEP)
            next_identifiers = filter_by_identifiers + identifiers_suffix
            # Because method must be inside allowed filter fields RFS, we must add value filter_by of the method
            # to the next RFS.
            next_filters_fields_rfs = rfs(filter_by_identifiers)
            return self._get_filter_recursive(
                identifiers_prefix, next_identifiers, model, resource, request, next_filters_fields_rfs
            )
        suffix = LOOKUP_SEP.join(identifiers_suffix)
        if not hasattr(method, 'filter') or (suffix and suffix not in method.filter.get_suffixes()):
            raise FilterIdentifierError
        return method.filter(identifiers_prefix, identifiers, identifiers_suffix, model, method=method)

    def _get_real_field_name(self, resource, field_name):
        return resource.renamed_fields.get(field_name, field_name) if resource else field_name

    def _get_resource_filter(self, identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs):
        """
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param filters_fields_rfs: RFS of fields that is allowed to filter.
        :return: method returns filter that is obtained from resource and its methods.
        """
        # Filter is obtained from resource filters dict
        for i in range(1, len(identifiers) + 1):
            # Because resource filters can contains filter key with __ we must try all combinations with suffixes
            current_identifiers = identifiers[:i]
            identifiers_string = self._get_real_field_name(resource, LOOKUP_SEP.join(current_identifiers))
            identifiers_suffix = identifiers[i:]
            suffix_string = LOOKUP_SEP.join(identifiers_suffix)
            if (resource and identifiers_string in resource.filters and
                    (not suffix_string or suffix_string in resource.filters[identifiers_string].get_suffixes())):
                return resource.filters[identifiers_string](
                    identifiers_prefix, current_identifiers, identifiers_suffix, model
                )

        # Filter is obtained from resource methods
        current_identifier = self._get_real_field_name(resource, identifiers[0])
        resource_method = resource.get_method_returning_field_value(current_identifier) if resource else None
        if current_identifier in filters_fields_rfs and resource_method:
            return self._get_method_filter(
                resource_method, identifiers_prefix, [current_identifier], identifiers[1:], model, resource, request,
                filters_fields_rfs
            )

        return None

    def _get_model_filter(self, identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs):
        """
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param filters_fields_rfs: RFS of fields that is allowed to filter.
        :return: method returns filter from model fields or methods
        """
        return None

    def _get_filters_fields_rfs(self, model, resource):
        return resource.get_filter_fields_rfs() if resource else rfs()

    def _get_filter_recursive(self, identifiers_prefix, identifiers, model, resource, request,
                              extra_filter_fields_rfs=None):
        """
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param extra_filter_fields_rfs: RFS of fields that is allowed to filter.
        :return: method search recursive filter with resource_filter and model_filter getters.
        """
        if not identifiers:
            return None

        extra_filter_fields_rfs = rfs() if extra_filter_fields_rfs is None else extra_filter_fields_rfs
        filters_fields_rfs = extra_filter_fields_rfs.join(self._get_filters_fields_rfs(model, resource))

        filter_obj = (
            self._get_resource_filter(
                identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs
            )
            or self._get_model_filter(
                identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs
            )
        )
        if not filter_obj:
            raise FilterIdentifierError
        return filter_obj

    def get_filter(self, identifiers, resource, request):
        """
        :param identifiers: list of identifiers that conclusively identifies a filter.
        :param resource: resource object.
        :param request: django HTTP request.
        :return: method returns filter object according to input identifiers, resource and request.
        """
        return self._get_filter_recursive([], identifiers, resource.model, resource, request)

    def filter(self, resource, qs, request):
        """
        :param resource: resource object.
        :param qs: model queryset that will be filtered.
        :param request: django HTTP request.
        :return: methods should return filtered queryset.
        """
        raise NotImplementedError


class BaseDjangoFilterManager(BaseModelFilterManager):
    """
    Filter manager is used inside model resource for composing filters with purpose to restrict output data according
    to input values.
    This is abstract class that provides methods to obtain concrete filters from resource and model methods and fields.
    """

    model_field_filters = {}

    def _get_default_model_field_filter_class(self, model_field):
        model_field_filter = getattr(model_field, 'filter', None)
        if model_field_filter:
            return model_field_filter

        for field_class, filter_class in list(self.model_field_filters.items())[::-1]:
            if isinstance(model_field, field_class):
                return filter_class

        return get_default_field_filter_class(model_field)

    def _get_model_filter(self, identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs):
        """
        :param identifiers_prefix: because filters are recursive if model relations property contains list of
               identifiers that was used for recursive searching the filter.
        :param identifiers: list of identifiers that conclusively identifies the filter.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param filters_fields_rfs: RFS of fields that is allowed to filter.
        :return: method returns filter from model fields or methods
        """
        current_identifier = self._get_real_field_name(resource, identifiers[0])
        identifiers_suffix = identifiers[1:]

        if current_identifier not in filters_fields_rfs:
            return None

        suffix = LOOKUP_SEP.join(identifiers_suffix)

        model_field = get_field_or_none(model, current_identifier)
        model_method = get_method_or_none(model, current_identifier)
        model_filter_filter = self._get_default_model_field_filter_class(model_field) if model_field else None

        if model_filter_filter and (not suffix or suffix in model_filter_filter.get_suffixes()):
            return model_filter_filter(
                identifiers_prefix, [current_identifier], identifiers_suffix, model, field=model_field
            )
        elif model_field and model_field.is_relation and model_field.related_model:
            # recursive search for filter via related model fields
            next_model = model_field.related_model
            next_resource = get_resource_or_none(request, next_model, getattr(resource, 'resource_typemapper', None))
            return self._get_filter_recursive(
                identifiers_prefix + [identifiers[0]], identifiers[1:], next_model, next_resource, request,
                filters_fields_rfs[current_identifier].subfieldset
            )
        elif model_method:
            return self._get_method_filter(
                model_method, identifiers_prefix, [current_identifier], identifiers_suffix,
                model, resource, request, filters_fields_rfs
            )
        else:
            return None

    def _get_filters_fields_rfs(self, model, resource):
        return resource.get_filter_fields_rfs() if resource else get_allowed_filter_fields_rfs_from_model(model)


def get_flat_lookups(q):
    if isinstance(q, Q):
        result = []
        for child in q.children:
            result += get_flat_lookups(child)
        return result
    else:
        return [q[0]]


def is_required_distinct_for_lookup(model, lookup, fail_first=False):
    if '__' in lookup:
        curr_lookup, next_lookup = lookup.split('__', 1)
    else:
        curr_lookup, next_lookup = lookup, None
    try:
        field = model._meta.get_field(curr_lookup)
        if not field.is_relation:
            return False
        elif field.many_to_many or field.one_to_many:
            # m2m or o2m relations can cause duplications
            return True
        elif not next_lookup:
            return False
        else:
            return is_required_distinct_for_lookup(field.related_model, next_lookup, True)
    except FieldDoesNotExist:
        # For annotations is distinct automatically required
        return fail_first


class BaseParserModelFilterManager(BaseModelFilterManager):

    parsers = [DefaultFilterParser(), QueryStringFilterParser()]

    def _logical_conditions_and(self, condition_a, condition_b):
        raise RestException(ugettext('More filter terms combination are not supported'))

    def _logical_conditions_or(self, condition_a, condition_b):
        raise RestException(ugettext('More filter terms combination are not supported'))

    def _logical_conditions_negation(self, condition):
        raise RestException(ugettext('Filter term negation are not supported'))

    def _convert_logical_conditions(self, condition, resource, request):
        """
        Method that recursive converts condition tree to the django models Q objects.
        """
        if condition.is_composed and condition.operator_slug == LOGICAL_OPERATORS.NOT:
            return self._logical_conditions_negation(
                self._convert_logical_conditions(condition.condition_right, resource, request)
            )
        elif condition.is_composed and condition.operator_slug == LOGICAL_OPERATORS.AND:
            return self._logical_conditions_and(
                self._convert_logical_conditions(condition.condition_left, resource, request),
                self._convert_logical_conditions(condition.condition_right, resource, request)
            )
        elif condition.is_composed and condition.operator_slug == LOGICAL_OPERATORS.OR:
            return self._logical_conditions_or(
                self._convert_logical_conditions(condition.condition_left, resource, request),
                self._convert_logical_conditions(condition.condition_right, resource, request)
            )
        else:
            try:
                return self.get_filter(condition.identifiers, resource, request).get_q(
                    condition.value, condition.operator_slug, request
                )
            except FilterIdentifierError:
                raise RestException(
                    mark_safe(ugettext('Invalid identifier of condition "{}"').format(condition.source))
                )
            except FilterValueError as ex:
                raise RestException(
                    mark_safe(ugettext('Invalid value of condition "{}". {}').format(condition.source, ex))
                )
            except OperatorFilterError:
                raise RestException(
                    mark_safe(ugettext('Invalid operator of condition "{}"').format(condition.source))
                )

    def _is_required_distinct(self, qs, q):
        for lookup in get_flat_lookups(q):
            if is_required_distinct_for_lookup(qs.model, lookup):
                return True
        return False

    def _filter_queryset(self, qs, q):
        raise NotImplementedError

    def _filter_with_parser(self, parser, resource, qs, request):
        try:
            parsed_conditions = parser.parse(request)
            if parsed_conditions:
                return self._filter_queryset(qs, self._convert_logical_conditions(parsed_conditions, resource, request))
            else:
                return qs
        except FilterParserError as ex:
            raise RestException(ex)

    def filter(self, resource, qs, request):
        for parser in self.parsers:
            qs = self._filter_with_parser(parser, resource, qs, request)
        return qs


class DjangoFilterManager(BaseParserModelFilterManager, BaseDjangoFilterManager):
    """
    Manager that uses parser to parse input filter data to the logical conditions tree.
    """

    def _logical_conditions_and(self, condition_a, condition_b):
        return Q(condition_a, condition_b)

    def _logical_conditions_or(self, condition_a, condition_b):
        return Q(condition_a | condition_b)

    def _logical_conditions_negation(self, condition):
        return ~Q(condition)

    def _is_required_distinct(self, qs, q):
        for lookup in get_flat_lookups(q):
            if is_required_distinct_for_lookup(qs.model, lookup):
                return True
        return False

    def _filter_queryset(self, qs, q):
        if self._is_required_distinct(qs, q):
            return qs.filter(pk__in=qs.filter(q).values('pk'))
        else:
            return qs.filter(q)



from django.test.utils import override_settings

from germanium.decorators import data_consumer
from germanium.tools.trivials import assert_in, assert_equal, assert_true, assert_is_not_none
from germanium.tools.http import (assert_http_bad_request, assert_http_not_found, assert_http_method_not_allowed,
                                  assert_http_accepted, build_url)
from germanium.tools.rest import assert_valid_JSON_created_response, assert_valid_JSON_response

from .factories import UserFactory, IssueFactory
from .test_case import PystonTestCase

from app.dynamo.models import Comment


class DynamodbTestCase(PystonTestCase):

    COMMENT_API_URL = '/api/issue/{}/Dynamo-comment/'

    @classmethod
    def setUpClass(cls):
        Comment.create_table()
        for i in range(10):
            comment = Comment(
                user_id=str(i),
                issue_id=str(i % 3),
                content=f'test message {i}',
                is_public=bool(i % 2),
                priority=i
            )
            comment.save()

    @classmethod
    def tearDownClass(cls):
        Comment.delete_table()

    def test_get_Dynamo_comments_should_return_right_data(self):
        resp = self.get(self.COMMENT_API_URL.format(0))
        assert_valid_JSON_response(resp)
        assert_equal(len(resp.json()), 4)
        for i, data in enumerate(resp.json()):
            assert_equal(data['priority'], i * 3)
            assert_equal(data['user_id'], str(i * 3))
            assert_equal(data['is_public'], bool(i % 2))
            assert_equal(data['content'], f'test message {i * 3}')
            assert_equal(data['issue_id'], str(0))

        resp = self.get(self.COMMENT_API_URL.format(1))
        assert_equal(len(resp.json()), 3)

        resp = self.get(self.COMMENT_API_URL.format(2))
        assert_equal(len(resp.json()), 3)

    def test_get_one_Dynamo_comment_should_return_right_data(self):
        resp = self.get(f'{self.COMMENT_API_URL.format(0)}3/')
        assert_valid_JSON_response(resp)
        data = resp.json()
        assert_equal(data['priority'], 3)
        assert_equal(data['user_id'], '3')
        assert_equal(data['is_public'], True)
        assert_equal(data['content'], f'test message 3')
        assert_equal(data['issue_id'], '0')

    def test_get_Dynamo_comments_should_be_sorted(self):
        resp = self.get(build_url(self.COMMENT_API_URL.format(0), order='user_id'))
        assert_equal([v['user_id'] for v in resp.json()], ['0', '3', '6', '9'])
        resp = self.get(build_url(self.COMMENT_API_URL.format(0), order='-user_id'))
        assert_equal([v['user_id'] for v in resp.json()], ['9', '6', '3', '0'])
        assert_http_bad_request(self.get(build_url(self.COMMENT_API_URL.format(0), order='id')))

    def test_get_Dynamo_comments_should_be_filtered(self):
        resp = self.get(build_url(self.COMMENT_API_URL.format(0), filter='user_id=0'))
        assert_equal(len(resp.json()), 1)
        resp = self.get(build_url(self.COMMENT_API_URL.format(0), filter='user_id=10'))
        assert_equal(len(resp.json()), 0)
        assert_http_bad_request(self.get(build_url(self.COMMENT_API_URL.format(0), filter='invalid=5')))

    def test_get_Dynamo__comments_should_paginated(self):
        headers = {'HTTP_X_BASE': '3'}
        resp = self.get(self.COMMENT_API_URL.format(0), headers=headers)
        assert_true('x-next-cursor' in resp)
        assert_equal([v['user_id'] for v in resp.json()], ['0', '3', '6'])
        headers['HTTP_X_CURSOR'] = resp['x-next-cursor']
        resp = self.get(self.COMMENT_API_URL.format(0), headers=headers)
        assert_equal([v['user_id'] for v in resp.json()], ['9'])




from django import forms
from django.db.models import F, Q
from pyston.converters import XmlConverter
from pyston.filters.filters import BooleanFilterMixin
from pyston.filters.django_filters import SimpleEqualFilter
from pyston.forms import (
    ISODateTimeField, MultipleRelatedField, RestModelForm, RestValidationError, ReverseManyField, ReverseOneToOneField,
    SingleRelatedField
)
from pyston.forms.postgres import RestSimpleArrayField
from pyston.resource import DjangoResource, BaseModelResource, BaseResource
from pyston.response import RestCreatedResponse
from pyston.serializer import SerializableObj

from .models import Issue, User
from .serializable import CountIssuesPerUserTable, CountWatchersPerIssue


class OvertimeIssuesFilter(BooleanFilterMixin, SimpleEqualFilter):

    def get_filter_term(self, value, operator, request):
        filter_term = Q(**{
            'solving_issue__in': Issue.objects.filter(logged_minutes__gt=F('estimate_minutes')).values('pk')
        })
        return filter_term if value else ~filter_term


class IssueResource(DjangoResource):

    model = Issue
    fields = ('id', 'created_at', '_obj_name', 'name', ('created_by', ('id', 'contract', 'created_at')), 'solver',
              'leader', 'watched_by', 'logged_minutes')
    detailed_fields = ('id', 'created_at', '_obj_name', 'name', ('created_by', ('id', 'contract',)), 'solver',
                       'leader', 'watched_by')
    general_fields = ('id', '_obj_name', 'name', ('created_by', ('id', 'contract', 'created_at')), 'watched_by',
                      'short_description')

    converter_classes = (
        'pyston.converters.JsonConverter',
        XmlConverter,
    )
    can_create_obj = True
    can_read_obj = True
    can_update_obj = True
    can_delete_obj = True

    def _obj_name(self, obj):
        return str(obj)


class UserResource(DjangoResource):

    model = User
    renamed_fields = {
        'createdAt': 'created_at',
        'solvingIssue': 'solving_issue',
        'firstName': 'first_name',
        'lastName': 'last_name',
        'isSuperuser': 'is_superuser',
        'watchedIssues': 'watched_issues',
        'createdIssues': 'created_issues',
        'manualCreatedDate': 'manual_created_date',
        'watchedIssuesCount': 'watched_issues_count',
    }
    fields = (
        'createdAt', 'email', 'contract', 'solvingIssue', 'firstName', 'lastName', 'isSuperuser', 'manualCreatedDate',
        'watchedIssuesCount'
    )
    detailed_fields = (
        'createdAt', 'email', 'contract', 'solvingIssue', 'firstName', 'lastName', 'watchedIssues',
        'createdIssues__id', 'manualCreatedDate'
    )
    general_fields = (
        'email', 'firstName', 'lastName', 'watchedIssues__name', 'watchedIssues__id', 'manualCreatedDate',
        'watchedIssuesCount'
    )
    extra_fields = ()
    can_create_obj = True
    can_read_obj = True
    can_update_obj = True
    can_delete_obj = True
    extra_order_fields = ()
    extra_filter_fields = ()
    filters = {
        'issues__overtime': OvertimeIssuesFilter
    }
    field_labels = {
        'email': 'E-mail address',
    }


class ExtraResource(BaseResource):

    def get(self):
        return {'extra': 1}


class CountIssuesPerUserResource(BaseResource):

    def get(self):
        return CountIssuesPerUserTable()


class CountWatchersPerIssueResource(BaseResource):

    def get(self):
        return [CountWatchersPerIssue(issue) for issue in Issue.objects.all()]


class TestTextObject(SerializableObj):

    def __init__(self, fiz_baz):
        self.fiz_baz = fiz_baz

    class Meta:
        fields = ('fiz_baz',)


class TestTextObjectCamelCaseResource(BaseModelResource):

    model = TestTextObject
    register = True

    can_read_obj = True

    renamed_fields = {
        'fizBaz': 'fiz_baz',
    }
    fields = ('fizBaz',)


class TestCamelCaseResource(BaseResource):

    def get(self):
        connected = TestTextObject('test object property content')
        return {
            'fooBar': 'foo bar',
            'connected': connected,
        }


class UserForm(RestModelForm):

    watched_issues = ReverseManyField('watched_issues')
    created_issues_renamed = ReverseManyField('created_issues')
    solving_issue_renamed = ReverseOneToOneField('solving_issue')
    leading_issue_renamed = ReverseOneToOneField('leading_issue')

    def clean_created_issues_renamed(self):
        created_issues = self.cleaned_data.get('created_issues_renamed')
        if created_issues and any(issue.name == 'invalid' for issue in created_issues):
            raise RestValidationError('Invalid issue name')


class UserWithFormResource(DjangoResource):

    register = False
    model = User
    form_class = UserForm
    can_create_obj = True
    can_read_obj = True
    can_update_obj = True
    can_delete_obj = True

    fields = ('user_email',)

    def user_email(self, obj):
        return obj.email


class IssueForm(RestModelForm):

    created_by = SingleRelatedField('created_by')
    leader = SingleRelatedField('leader', is_allowed_foreign_key=False)
    another_users = MultipleRelatedField('watched_by', form_field=forms.ModelMultipleChoiceField(
        queryset=User.objects.all(), required=False
    ), is_allowed_foreign_key=False)
    tags_list = RestSimpleArrayField(label='tags', base_field=forms.CharField(max_length=5), required=False)
    iso_date = ISODateTimeField(required=False)

    def save(self, commit=True):
        instance = super().save(commit=False)
        tags_list = self.cleaned_data.get('tags_list')
        if tags_list:
            instance.tags = '|'.join(tags_list)
        if commit:
            instance.save()
        return instance

    class Meta:
        exclude = ('tags',)


class IssueWithFormResource(DjangoResource):

    register = False
    model = Issue
    form_class = IssueForm
    can_create_obj = True
    can_read_obj = True
    can_update_obj = True
    can_delete_obj = True

    fields = ('id', 'creator')
    resource_typemapper = {
        User: UserWithFormResource,
    }

    def creator(self, obj):
        return obj.created_by



from django.db import models
from django.db.models import Count
from django.utils.translation import ugettext_lazy as _

from pyston.utils.decorators import order_by, filter_class, filter_by, allow_tags, sorter_class
from pyston.filters.filters import OPERATORS, IntegerFilterMixin
from pyston.filters.django_filters import CONTAINS, StringFieldFilter, SimpleMethodEqualFilter
from pyston.order.django_sorters import ExtraDjangoSorter


class OnlyContainsStringFieldFilter(StringFieldFilter):

    operators = (
        (OPERATORS.CONTAINS, CONTAINS),
    )


class WatchedIssuesCountMethodFilter(IntegerFilterMixin, SimpleMethodEqualFilter):

    def get_filter_term(self, value, operator, request):
        return {
            'pk__in': User.objects.annotate(
                watched_issues_count=Count('watched_issues')
            ).filter(watched_issues_count=value).values('pk')
        }


class WatchedIssuesCountSorter(ExtraDjangoSorter):

    def update_queryset(self, qs):
        return qs.annotate(**{self._get_order_string(): Count('watched_issues')})


class User(models.Model):

    created_at = models.DateTimeField(verbose_name=_('created at'), null=False, blank=False, auto_now_add=True)
    email = models.EmailField(verbose_name=_('email'), null=False, blank=False, unique=True,
                              filter=OnlyContainsStringFieldFilter)
    contract = models.FileField(_('file'), null=True, blank=True, upload_to='documents/')
    is_superuser = models.BooleanField(_('is superuser'), default=True)
    first_name = models.CharField(_('first name'), null=True, blank=True, max_length=100)
    last_name = models.CharField(_('last name'), null=True, blank=True, max_length=100)
    manual_created_date = models.DateTimeField(verbose_name=_('manual created date'), null=True, blank=True)

    @filter_class(WatchedIssuesCountMethodFilter)
    @sorter_class(WatchedIssuesCountSorter)
    def watched_issues_count(self):
        return self.watched_issues.count()

    def __str__(self):
        return 'user: %s' % self.email

    class RestMeta:
        fields = ('created_at', 'email', 'contract', 'solving_issue', 'first_name', 'last_name', 'is_superuser',
                  'manual_created_date', 'watched_issues_count')
        detailed_fields = ('created_at', '_obj_name', 'email', 'contract', 'solving_issue', 'first_name', 'last_name',
                           'watched_issues__name', 'watched_issues__id', 'manual_created_date')
        general_fields = ('email', 'first_name', 'last_name', 'watched_issues__name', 'watched_issues__id',
                          'manual_created_date', 'watched_issues_count')
        direct_serialization_fields = ('created_at', 'email', 'contract', 'solving_issue', 'first_name', 'last_name',
                                       'manual_created_date')
        order_fields = ('email', 'solving_issue', 'watched_issues_count')
        extra_order_fields = ('created_at',)
        filter_fields = ('email', 'first_name', 'last_name', 'is_superuser', 'created_issues', 'watched_issues_count')
        extra_filter_fields = ('created_at',)


class Issue(models.Model):

    created_at = models.DateTimeField(verbose_name=_('created at'), null=False, blank=False, auto_now_add=True)
    name = models.CharField(verbose_name=_('name'), max_length=100, null=False, blank=False)
    watched_by = models.ManyToManyField('app.User', verbose_name=_('watched by'), blank=True,
                                        related_name='watched_issues')
    created_by = models.ForeignKey('app.User', verbose_name=_('created by'), null=False, blank=False,
                                   related_name='created_issues', on_delete=models.CASCADE)
    solver = models.OneToOneField('app.User', verbose_name=_('solver'), null=True, blank=True,
                                  related_name='solving_issue', on_delete=models.CASCADE)
    leader = models.OneToOneField('app.User', verbose_name=_('leader'), null=False, blank=False,
                                  related_name='leading_issue', on_delete=models.CASCADE)
    description = models.TextField(verbose_name=_('description'), null=True, blank=True)
    logged_minutes = models.IntegerField(verbose_name=_('logged minutes'), null=True, blank=True)
    estimate_minutes = models.IntegerField(verbose_name=_('logged minutes'), null=True, blank=True)
    tags = models.TextField(verbose_name=_('tags'), null=True, blank=True)

    @filter_by('description')
    @order_by('description')
    @allow_tags
    def short_description(self):
        return self.description[:50] if self.description is not None else None

    def __str__(self):
        return 'issue: <b>%s</b>' % self.name

    class RestMeta:
        extra_order_fields = ('solver__created_at',)
        extra_filter_fields = ('solver__created_at',)


from django.core.exceptions import ValidationError
from django.contrib.postgres.utils import prefix_validation_error
from django.contrib.postgres.forms.array import SimpleArrayField
from django.utils.translation import ugettext


class RestSimpleArrayField(SimpleArrayField):
    """
    Django array form field doesn't accept list. Therefore we must rewrite to_python method
    """

    def to_python(self, value):
        if value is None:
            return None

        if isinstance(value, list):
            errors = []
            values = []
            for index, item in enumerate(value):
                try:
                    values.append(self.base_field.to_python(item))
                except ValidationError as error:
                    errors.append(prefix_validation_error(
                        error,
                        prefix=self.error_messages['item_invalid'],
                        code='item_invalid',
                        params={'nth': index},
                    ))
            if errors:
                raise ValidationError(errors)
        else:
            raise ValidationError(ugettext('Enter a list.'))
        return values

    def clean(self, value):
        if value is None:
            return value
        else:
            return super().clean(value)



from pyston.contrib.elasticsearch.resource import BaseElasticsearchResource

from .models import Comment


class CommentElasticsearchResource(BaseElasticsearchResource):

    model = Comment
    fields = (
        'id', 'user_id', 'content', 'is_public', 'priority'
    )

    can_read_obj = True



from django.utils.translation import ugettext_lazy as _


class UnsupportedMediaTypeException(Exception):
    """
    Raised if the content_type has unsupported media type
    """
    pass


class MimerDataException(Exception):
    """
    Raised if the content_type and data don't match
    """
    pass


class RestException(Exception):
    message = None

    def __init__(self, message=None):
        super().__init__()
        self.message = message or self.message

    @property
    def errors(self):
        return {'error': self.message}


class UnprocessableEntity(RestException):
    pass


class ResourceNotFoundException(RestException):
    message = _('Select a valid choice. That choice is not one of the available choices.')


class NotAllowedException(RestException):
    message = _('Not allowed.')


class UnauthorizedException(RestException):
    message = _('Unauthorized.')


class NotAllowedMethodException(RestException):
    message = _('Not allowed method.')


class DuplicateEntryException(RestException):
    message = _('Conflict/Duplicate.')


class ConflictException(RestException):
    message = _('Object already exists but you do not allowed to change it.')


class DataInvalidException(Exception):

    def __init__(self, errors):
        self.message = errors
        self.errors = errors



__all__ = (
    'allow_tags',
    'humanized',
    'filter_class',
    'filter_by',
    'order_by',
    'sorter_class',
)


def allow_tags(func):
    """Allows HTML tags to be returned from resource without escaping"""
    if isinstance(func, property):
        func = func.fget
    func.allow_tags = True
    return func


def humanized(humanized_func, **humanized_func_kwargs):
    """Sets 'humanized' function to method or property."""
    def decorator(func):
        if isinstance(func, property):
            func = func.fget

        def _humanized_func(*args, **kwargs):
            kwargs.update(humanized_func_kwargs)
            return humanized_func(*args, **kwargs)
        func.humanized = _humanized_func
        return func
    return decorator


def filter_class(filter_class):
    """Sets 'filter' class (this attribute is used inside grid and rest)."""
    def decorator(func):
        if isinstance(func, property):
            func = func.fget
        func.filter = filter_class
        return func
    return decorator


def filter_by(field_name):
    """Sets 'field name' (this is used for grid filtering)"""
    def decorator(func):
        if isinstance(func, property):
            func = func.fget
        func.filter_by = field_name
        return func
    return decorator


def order_by(field_name):
    """Sets 'field name' (this is used for grid ordering)"""
    def decorator(func):
        if isinstance(func, property):
            func = func.fget
        func.order_by = field_name
        return func
    return decorator


def sorter_class(sorter_class):
    """Sets 'sorter' class (this attribute is used inside grid and rest)."""
    def decorator(func):
        if isinstance(func, property):
            func = func.fget
        func.sorter = sorter_class
        return func
    return decorator



import base64
import json

from django.utils.translation import ugettext

from pyston.exception import RestException
from pyston.paginator import BasePaginator
from pyston.response import HeadersResponse


class DynamoCursorBasedPaginator(BasePaginator):

    def __init__(self, default_base=20, max_base=100):
        self.default_base = default_base
        self.max_base = max_base

    def get_response(self, qs, request):
        base = self._get_base(request)
        cursor = self._get_cursor(request)

        if cursor is not None:
            qs = qs.set_last_evaluated_key(cursor)
        qs = qs.set_limit(base)

        return HeadersResponse(
            qs, self.get_headers(self.get_next_cursor(qs))
        )

    def _get_base(self, request):
        base = request._rest_context.get('base')
        if not base:
            return self.default_base
        elif base.isdigit():
            base_int = int(base)
            if base_int > self.max_base:
                raise RestException(ugettext('Base must lower or equal to {}').format(self.max_base))
            else:
                return base_int
        else:
            raise RestException(ugettext('Base must be natural number or empty'))

    def _get_cursor(self, request):
        cursor = request._rest_context.get('cursor')
        if cursor:
            try:
                cursor = base64.b64decode(
                    cursor.encode('ascii')
                ).decode('ascii')
                return json.loads(cursor)
            except json.JSONDecodeError:
                raise RestException(ugettext('Invalid next cursor value'))
        else:
            return None

    def get_next_cursor(self, qs):
        return base64.b64encode(
            json.dumps(qs.next_key).encode('ascii')
        ).decode('ascii') if qs.next_key else None

    def get_headers(self, next_cursor):
        return {
            k: v for k, v in {
                'X-Next-Cursor': next_cursor,
            }.items() if v is not None
        }



from pyston.resource import BaseModelResource
from pyston.serializer import ModelResourceSerializer, ModelSerializer, Serializer, register

from pydjamodb.queryset import DynamoDBQuerySet
from pynamodb.attributes import MapAttribute
from pynamodb.models import Model
from pynamodb.pagination import ResultIterator

from .order import DynamoOrderManager
from .filter import DynamoFilterManager
from .paginator import DynamoCursorBasedPaginator


@register(MapAttribute)
class MapAttributeSerializer(Serializer):

    def serialize(self, data, serialization_format, **kwargs):
        return {
            k: self._data_to_python(v, serialization_format, **kwargs)
            for k, v in data.attribute_values.items()
        }


@register((Model, ResultIterator))
class DynamoSerializer(ModelSerializer):

    obj_class = Model
    obj_iterable_classes = (ResultIterator, DynamoDBQuerySet)


class DynamoResourceSerializer(ModelResourceSerializer, DynamoSerializer):

    obj_iterable_classes = (ResultIterator, DynamoDBQuerySet)


class BaseDynamoResource(BaseModelResource):

    register = True
    abstract = True

    allowed_methods = ('get', 'head', 'options')
    paginator = DynamoCursorBasedPaginator()
    serializer = DynamoResourceSerializer
    order_manager = DynamoOrderManager()
    filter_manager = DynamoFilterManager()

    range_key = None

    def _get_hash_key(self):
        raise NotImplementedError

    def _get_obj_or_none(self, pk=None):
        if not pk:
            return None
        try:
            return self.model.get(self._get_hash_key(), pk)
        except self.model.DoesNotExist:
            return None

    def _get_queryset(self):
        return self.model.objects.set_hash_key(self._get_hash_key())

    def get_range_key(self):
        return self.range_key



from pyston.order.sorters import BaseSorter
from pyston.order.managers import BaseParserModelOrderManager
from pyston.order.utils import DIRECTION


class DynamoSorter(BaseSorter):

    def get_order_term(self):
        return self.direction == DIRECTION.ASC


class DynamoOrderManager(BaseParserModelOrderManager):

    def _sort_queryset(self, qs, terms):
        for term in terms:
            qs = qs.set_scan_index_forward(term)
        return qs

    def _get_sorter_from_model(self, identifiers_prefix, identifiers, direction, model, resource, request,
                               order_fields_rfs):
        if len(identifiers) == 1:
            current_identifier = self._get_real_field_name(resource, identifiers[0])
            if current_identifier == resource.get_range_key():
                return DynamoSorter(identifiers_prefix + identifiers, direction)
        return None



import types
import sys

from io import BytesIO

import datetime
import decimal
import uuid

from django.utils.encoding import force_bytes
from django.conf import settings
from django.utils.duration import duration_iso_string
from django.utils.functional import Promise
from django.utils.timezone import is_aware

from chamber.utils import get_class_method

from .compatibility import FieldDoesNotExist


class ModelIteratorHelper:

    def __init__(self, model):
        self.model = model

    def __iter__(self):
        raise NotImplementedError


class ModelIterableIteratorHelper(ModelIteratorHelper):

    def __init__(self, iterable, model):
        super().__init__(model)
        self.iterable = iterable

    def __iter__(self):
        return iter(self.iterable)


class UniversalBytesIO:

    def __init__(self, container=None, charset=None):
        self.charset = charset or settings.DEFAULT_CHARSET
        self._container = BytesIO() if container is None else container

    # These methods partially implement the file-like object interface.
    # See https://docs.python.org/3/library/io.html#io.IOBase

    def close(self):
        self._container.close()

    def write(self, content):
        self._container.write(self.make_bytes(content))

    def flush(self):
        self._container.flush()

    def tell(self):
        return self._container.tell()

    def readable(self):
        return False

    def seekable(self):
        return False

    def writable(self):
        return True

    def writelines(self, lines):
        for line in lines:
            self.write(line)

    def make_bytes(self, value):
        """Turn a value into a bytestring encoded in the output charset."""
        if isinstance(value, bytes):
            return bytes(value)
        if isinstance(value, str):
            return bytes(value.encode(self.charset))

        # Handle non-string types
        return force_bytes(value, self.charset)

    def getvalue(self):
        return self._container.getvalue().decode(self.charset)

    if sys.version_info[0:2] < (3, 5):
        def seek(self, *args, **kwargs):
            pass


def serialized_data_to_python(data):
    from pyston.serializer import LAZY_SERIALIZERS

    if isinstance(data, (types.GeneratorType, list, tuple)):
        return [serialized_data_to_python(val) for val in data]
    elif isinstance(data, LAZY_SERIALIZERS):
        return serialized_data_to_python(data.serialize())
    elif isinstance(data, dict):
        return {key: serialized_data_to_python(val) for key, val in data.items()}
    if isinstance(data, datetime.datetime):
        r = data.isoformat()
        if data.microsecond:
            r = r[:23] + r[26:]
        if r.endswith('+00:00'):
            r = r[:-6] + 'Z'
        return r
    elif isinstance(data, datetime.date):
        return data.isoformat()
    elif isinstance(data, datetime.time):
        if is_aware(data):
            raise ValueError("JSON can't represent timezone-aware times.")
        r = data.isoformat()
        if data.microsecond:
            r = r[:12]
        return r
    elif isinstance(data, datetime.timedelta):
        return duration_iso_string(data)
    elif isinstance(data, (decimal.Decimal, uuid.UUID, Promise)):
        return str(data)
    else:
        return data


def str_to_class(class_string):
    module_name, class_name = class_string.rsplit('.', 1)
    # load the module, will raise ImportError if module cannot be loaded
    m = __import__(module_name, globals(), locals(), str(class_name))
    # get the class, will raise AttributeError if class cannot be found
    c = getattr(m, class_name)
    return c


def get_field_or_none(model, field_name):
    try:
        return model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return None


def get_method_or_none(model, name):
    try:
        return get_class_method(model, name)
    except (AttributeError, FieldDoesNotExist):
        return None



from functools import reduce
from operator import or_

from chamber.shortcuts import get_object_or_none

from django.db.models import Q
from django.db.models.expressions import OrderBy
from django.utils.translation import ugettext

from .forms import RestValidationError
from .exception import RestException
from .utils.compatibility import get_last_parent_pk_field_name
from .utils.helpers import ModelIterableIteratorHelper
from .response import HeadersResponse


def _get_attr(obj, attr):
    if '__' in attr:
        rel_obj, rel_attr = attr.split('__')
        return _get_attr(getattr(obj, rel_obj), rel_attr)
    else:
        return getattr(obj, attr)


class BasePaginator:

    def get_response(self, qs, request):
        raise NotImplementedError


class BaseModelOffsetBasedPaginator(BasePaginator):

    iterable_helper_class = ModelIterableIteratorHelper

    def __init__(self, max_offset=pow(2, 63) - 1, max_base=100, default_base=20):
        self.max_offset = max_offset
        self.max_base = max_base
        self.default_base = default_base

    def _get_model(self, qs):
        raise NotImplementedError

    def _get_list_from_queryset(self, qs, from_, to_):
        raise NotImplementedError

    def _get_total(self, qs, request):
        raise NotImplementedError

    def _get_offset(self, qs, request):
        offset = request._rest_context.get('offset', '0')
        if offset.isdigit():
            offset_int = int(offset)
            if offset_int > self.max_offset:
                raise RestException(ugettext('Offset must be lower or equal to {}').format(self.max_offset))
            else:
                return offset_int
        else:
            raise RestException(ugettext('Offset must be natural number'))

    def _get_base(self, qs, request):
        base = request._rest_context.get('base')
        if not base:
            return self.default_base
        elif base.isdigit():
            base_int = int(base)
            if base_int > self.max_base:
                raise RestException(ugettext('Base must lower or equal to {}').format(self.max_base))
            else:
                return base_int
        else:
            raise RestException(ugettext('Base must be natural number or empty'))

    def _get_next_offset(self, iterable, offset, base):
        return offset + base if len(iterable) > base else None

    def _get_prev_offset(self, iterable, offset, base):
        return None if offset == 0 or not base else max(offset - base, 0)

    def _get_headers(self, total, next_offset, prev_offset):
        return {
            k: v for k, v in {
                'X-Total': total,
                'X-Next-Offset': next_offset,
                'X-Prev-Offset': prev_offset,
            }.items() if v is not None
        }

    def get_response(self, qs, request):
        base = self._get_base(qs, request)
        total = self._get_total(qs, request)
        offset = self._get_offset(qs, request)
        model = self._get_model(qs)
        # To check next offset, one more object is get from queryset
        iterable = self._get_list_from_queryset(qs, offset, offset + base + 1)
        next_offset = self._get_next_offset(iterable, offset, base)
        prev_offset = self._get_prev_offset(iterable, offset, base)
        return HeadersResponse(
            self.iterable_helper_class(iterable[:base], model),
            self._get_headers(total, next_offset, prev_offset)
        )


class DjangoOffsetBasedPaginator(BaseModelOffsetBasedPaginator):
    """
    REST paginator for list and querysets
    """

    def _get_model(self, qs):
        return qs.model

    def _get_list_from_queryset(self, qs, from_, to_):
        return list(qs[from_:to_])

    def _get_total(self, qs, request):
        return qs.count()


class DjangoOffsetBasedPaginatorWithoutTotal(DjangoOffsetBasedPaginator):

    def _get_total(self, qs, request):
        return None


class CursorBasedModelIterableIteratorHelper(ModelIterableIteratorHelper):

    def __init__(self, iterable, model, next):
        super().__init__(iterable, model)
        self.next = next


class DjangoCursorBasedPaginator(BasePaginator):

    def __init__(self, max_base=100, default_base=20):
        self.max_base = max_base
        self.default_base = default_base

    def get_response(self, qs, request):
        base = self._get_base(request)
        cursor = self._get_cursor(request)

        ordering = self._get_ordering(request, qs)

        cursor_based_model_iterable = self._get_paged_qs(qs, ordering, cursor, base)

        return HeadersResponse(
            cursor_based_model_iterable,
            self.get_headers(cursor_based_model_iterable.next)
        )

    def _get_page_filter_kwargs(self, current_row, ordering):
        ordering = list(ordering)

        args_or = []
        while ordering:
            base_order_field_name = ordering.pop()
            is_reverse = base_order_field_name.startswith('-')
            base_order_field_name = self._get_field_name(base_order_field_name)
            base_order_filtered_value = _get_attr(current_row, base_order_field_name)
            if base_order_filtered_value is None:
                if is_reverse:
                    filter_lookup = Q(**{'{}__isnull'.format(base_order_field_name): False})
                else:
                    # skip this filter
                    continue
            else:
                if is_reverse:
                    filter_lookup = Q(
                        **{'{}__lt'.format(base_order_field_name): base_order_filtered_value}
                    )
                else:
                    filter_lookup = Q(
                        **{'{}__gt'.format(base_order_field_name): base_order_filtered_value}
                    ) | Q(
                        **{'{}__isnull'.format(base_order_field_name): True}
                    )

            args_or.append(
                Q(
                    filter_lookup,
                    Q(**{
                        self._get_field_name(order): _get_attr(
                            current_row, self._get_field_name(order)
                        ) for order in ordering
                    })
                )
            )
        return reduce(or_, args_or)

    def _get_page(self, qs, base):
        results = list(qs[:base + 1])
        page = list(results[:base])
        next_cursor = self._get_position_from_instance(page[-1]) if len(results) > len(page) else None
        return CursorBasedModelIterableIteratorHelper(page, qs.model, next=next_cursor)

    @property
    def _get_paged_qs(self, qs, ordering, cursor, base):
        qs = qs.order_by(*ordering)

        if cursor:
            current_row = get_object_or_none(qs, pk=cursor)
            if current_row:
                qs = qs.filter(self._get_page_filter_kwargs(current_row))
            else:
                raise RestException(RestValidationError(ugettext('Cursor object was not found')))
        return self._get_page(qs, base)

    def _get_base(self, request):
        base = request._rest_context.get('base')
        if not base:
            return self.default_base
        elif base.isdigit():
            base_int = int(base)
            if base_int > self.max_base:
                raise RestException(ugettext('Base must lower or equal to {}').format(self.max_base))
            else:
                return base_int
        else:
            raise RestException(ugettext('Base must be natural number or empty'))

    def _get_cursor(self, request):
        return request._rest_context.get('cursor')

    def _get_ordering(self, request, qs):
        pk_field_name = get_last_parent_pk_field_name(qs.model)
        query_ordering = list(qs.query.order_by) or list(qs.model._meta.ordering)

        ordering = []
        for order_lookup in query_ordering:
            if isinstance(order_lookup, OrderBy):
                ordering.append(
                    '-' + order_lookup.expression.name if order_lookup.descending else order_lookup.expression.name
                )
            else:
                ordering.append(order_lookup)

        if self._pk_field_name not in ordering:
            ordering.append(pk_field_name)

        return ordering

    def _get_position_from_instance(self, instance):
        pk_field_name = get_last_parent_pk_field_name(instance.__class__)
        if isinstance(instance, dict):
            attr = instance[pk_field_name]
        else:
            attr = getattr(instance, pk_field_name)
        return str(attr)

    def _get_field_name(self, order_lookup):
        return order_lookup[1:] if order_lookup.startswith('-') else order_lookup

    def get_headers(self, next_cursor):
        return {
            k: v for k, v in {
                'X-Next-Cursor': next_cursor,
            }.items() if v is not None
        }



from pyston.exception import RestException

from .parser import DefaultRequestedFieldsParser


class ModelRequestedFieldsManager:

    parser = None

    def get_requested_fields(self, resource, request):
        try:
            parsed_requested_rfs = self.parser.parse(request)
            if parsed_requested_rfs is None:
                return None
            else:
                return parsed_requested_rfs
        except RestException as ex:
            raise RestException(ex)


class DefaultRequestedFieldsManager(ModelRequestedFieldsManager):
    """
    Default requested fields manager.
    """

    parser = DefaultRequestedFieldsParser()



from django.conf import settings
from .fields import *
from .fieldsets import *
from .filter import *
from .order import *
from .standard_operations import *
from .extra_resource import *
from .cors import *
from .standard_operations import *
from .compatibility import *
from .serializer import *
from .elasticsearch import *
from .dynamo import *



from pyston.resource import BaseModelResource
from pyston.serializer import ModelResourceSerializer, ModelSerializer, register
from pyston.utils.helpers import ModelIteratorHelper

from elasticsearch import NotFoundError
from elasticsearch_dsl import Document, Search

from .filters import ElasticsearchFilterManager
from .order import ElasticsearchOrderManager
from .paginator import ElasticsearchOffsetBasedPaginator


@register((Document, Search))
class ElasticsearchSerializer(ModelSerializer):

    obj_class = Document
    obj_iterable_classes = (ModelIteratorHelper, Search)

    def _field_to_python(self, field_name, real_field_name, obj, serialization_format, allow_tags=False, **kwargs):
        if field_name in obj._doc_type.mapping.properties:
            return self._data_to_python(
                self._value_to_raw_verbose(
                    getattr(obj, real_field_name),
                    obj,
                    allow_tags=False,
                    serialization_format=serialization_format,
                    **kwargs
                ),
                serialization_format,
                **kwargs
            )
        else:
            return super()._field_to_python(field_name, real_field_name, obj, serialization_format, allow_tags,
                                            **kwargs)


class ElasticsearchResourceSerializer(ModelResourceSerializer, ElasticsearchSerializer):

    obj_iterable_classes = (ModelIteratorHelper, Search)


class BaseElasticsearchResource(BaseModelResource):

    register = True
    abstract = True

    allowed_methods = ('get', 'head', 'options')
    serializer = ElasticsearchResourceSerializer
    paginator = ElasticsearchOffsetBasedPaginator()
    order_manager = ElasticsearchOrderManager()
    filter_manager = ElasticsearchFilterManager()

    def _get_queryset(self):
        return self.model.search()

    def _get_obj_or_none(self, pk=None):
        if not pk:
            return None
        try:
            return self.model.get(id=pk)
        except NotFoundError:
            return None



from pyston.paginator import BaseModelOffsetBasedPaginator


class ElasticsearchOffsetBasedPaginator(BaseModelOffsetBasedPaginator):

    def _get_total(self, qs, request):
        return qs.count()

    def _get_list_from_queryset(self, qs, from_, to_):
        return list(qs[from_:to_])

    def _get_model(self, qs):
        return qs._doc_type



from pyston.filters.filters import Filter, OPERATORS, BooleanFilterMixin, DateFilterMixin
from pyston.filters.exceptions import OperatorFilterError
from pyston.filters.managers import BaseParserModelFilterManager

from elasticsearch_dsl import Q


class ElasticsearchFilter(Filter):

    allowed_operators = None

    def get_q(self, value, operator_slug, request):
        if operator_slug not in self.get_allowed_operators():
            raise OperatorFilterError
        else:
            return self.get_filter_term(self.clean_value(value, operator_slug, request), operator_slug, request)

    def get_filter_term(self, value, operator_slug, request):
        """
        :return: returns Q object or dictionary that will be used for filtering resource response.
        """
        raise NotImplementedError


class WildcardElasticsearchFilter(ElasticsearchFilter):

    allowed_operators = (OPERATORS.CONTAINS, OPERATORS.EQ)

    def clean_value(self, value, operator_slug, request):
        if operator_slug == OPERATORS.EQ:
            return value
        else:
            return '*{}*'.format(value.replace('*', '\\*').replace('?', '\\?'))

    def get_filter_term(self, value, operator_slug, request):
        return Q('term' if operator_slug == OPERATORS.EQ else 'wildcard', **{self.get_full_filter_key(): value})


class MatchElasticsearchFilter(ElasticsearchFilter):

    allowed_operators = (OPERATORS.ICONTAINS,)

    def get_filter_term(self, value, operator_slug, request):
        return Q('match_phrase', **{self.get_full_filter_key(): value})


class BooleanElasticsearchFilter(BooleanFilterMixin, ElasticsearchFilter):

    allowed_operators = (OPERATORS.EQ,)

    def get_filter_term(self, value, operator_slug, request):
        return Q('term', **{self.get_full_filter_key(): value})


class IDElasticsearchFilter(ElasticsearchFilter):

    allowed_operators = (OPERATORS.EQ,)

    def get_filter_term(self, value, operator_slug, request):
        return Q('ids', **{'values': [value]})


class DateTimeElasticsearchFilter(DateFilterMixin, ElasticsearchFilter):

    suffixes = {}
    allowed_operators = (OPERATORS.GT, OPERATORS.GTE, OPERATORS.LT, OPERATORS.LTE)

    def get_filter_term(self, value, operator_slug, request):
        return Q('range', **{self.get_full_filter_key(): {operator_slug: value}})


class ElasticsearchFilterManager(BaseParserModelFilterManager):

    filter_by_field_name = {
        'text': MatchElasticsearchFilter,
        'jsontext': MatchElasticsearchFilter,
        'keyword': WildcardElasticsearchFilter,
        'boolean': BooleanElasticsearchFilter,
        'date': DateTimeElasticsearchFilter,
        'id': IDElasticsearchFilter,
    }

    def _logical_conditions_and(self, condition_a, condition_b):
        return condition_a & condition_b

    def _logical_conditions_or(self, condition_a, condition_b):
        return condition_a | condition_b

    def _logical_conditions_negation(self, condition):
        return ~condition

    def _get_model_filter(self, identifiers_prefix, identifiers, model, resource, request, filters_fields_rfs):
        current_identifier = self._get_real_field_name(resource, identifiers[0])
        identifiers_suffix = identifiers[1:]

        if current_identifier in filters_fields_rfs and not identifiers_prefix and not identifiers_suffix:
            try:
                if current_identifier == 'id':
                    field = None
                    field_name = 'id'
                else:
                    field = model._doc_type.mapping.properties[current_identifier]
                    field_name = field.name
                    if hasattr(field, 'builtin_type') and field_name not in self.filter_by_field_name:
                        field_name = field.builtin_type.name

                return self.filter_by_field_name[field_name](identifiers_prefix, identifiers, [], model, field=field)
            except KeyError:
                return None

        return None

    def _filter_queryset(self, qs, q):
        return qs.filter(q)



import binascii
import sys
import base64
import inspect

from io import BytesIO

from django.forms.fields import FileField
from django.utils.translation import ugettext_lazy as _, ugettext
from django.core.files.uploadedfile import InMemoryUploadedFile
from django.core.validators import URLValidator
from django.core.exceptions import ValidationError

from requests.exceptions import RequestException

from pyston.conf import settings as pyston_settings
from pyston.utils.files import (
    get_file_name_type_and_content_from_url, get_content_type_from_filename, get_content_type_from_file_content,
    get_filename_from_content_type, RequestDataTooBig, InvalidResponseStatusCode
)

from .exception import DataInvalidException
from .resource import BaseModelResource, DjangoResource
from .forms import RestDictError, RestError, RestValidationError


url_validator = URLValidator()


class DataProcessorCollection:

    def __init__(self):
        self.data_processors_map = {}

    def register(self, resource_class):
        def _register(processor_class):
            data_processors = self.data_processors_map.get(resource_class, set())
            data_processors.add(processor_class)
            self.data_processors_map[resource_class] = data_processors
            return processor_class
        return _register

    def get_processors(self, resource_class):
        processors = []
        for obj_class in inspect.getmro(resource_class):
            processors += list(self.data_processors_map.get(obj_class, set()))
        return processors


data_preprocessors = DataProcessorCollection()
data_postprocessors = DataProcessorCollection()


class DataProcessor:

    def __init__(self, resource, form, *args, **kwargs):
        self.resource = resource
        self.request = resource.request
        self.form = form

    def _process_field(self, data, files, key, data_item):
        raise NotImplementedError

    def _clear_data(self, data, files):
        return data, files

    def process_data(self, data, files):
        data, files = self._clear_data(data, files)

        self.errors = RestDictError()
        for key, data_item in data.items():
            self._process_field(data, files, key, data_item)

        if self.errors:
            raise DataInvalidException(self.errors)
        return data, files


@data_preprocessors.register(BaseModelResource)
class FileDataPreprocessor(DataProcessor):

    def _validate_not_empty(self, data_item, key, item):
        if not data_item.get(item):
            error = self.errors.get(key, RestDictError())
            error.update(RestDictError({item: RestValidationError(ugettext('This field is required'))}))
            self.errors[key] = error

    def _get_content_type(self, content_type, filename, file_content):
        if content_type:
            return content_type
        elif filename:
            return get_content_type_from_filename(filename)
        else:
            return get_content_type_from_file_content(file_content)

    def _get_filename(self, content_type, filename):
        if filename:
            return filename
        else:
            return get_filename_from_content_type(content_type)

    def _process_file_data(self, data, files, key, data_item, file_content):
        filename = data_item.get('filename')
        content_type = data_item.get('content_type')

        content_type = self._get_content_type(content_type, filename, file_content)
        filename = self._get_filename(content_type, filename)
        if not content_type:
            self.errors[key] = RestValidationError(ugettext(
                'Content type cannot be evaluated from input data please send it'
            ))
        elif not filename:
            self.errors[key] = RestValidationError(ugettext(
                'File name cannot be evaluated from input data please send it'
            ))
        else:
            filename = self._get_filename(content_type, filename)
            charset = data_item.get('charset')
            files[key] = InMemoryUploadedFile(
                file_content, field_name=key, name=filename, content_type=content_type,
                size=sys.getsizeof(file_content), charset=charset
            )
            data[key] = filename

    def _process_file_data_field(self, data, files, key, data_item):
        try:
            file_content = BytesIO(base64.b64decode(data_item.get('content').encode('utf-8')))
            self._process_file_data(data, files, key, data_item, file_content)
        except (TypeError, binascii.Error):
            self.errors[key] = RestDictError({'content': RestValidationError(
                ugettext('File content must be in base64 format')
            )})

    def _process_file_data_url_field(self, data, files, key, data_item):
        url = data_item.get('url')
        try:
            file_name, content_type, file_content = get_file_name_type_and_content_from_url(
                url, pyston_settings.FILE_SIZE_LIMIT
            )

            data_item['filename'] = data_item.get('filename', file_name)
            data_item['content_type'] = data_item.get('content_type', content_type)

            self._process_file_data(data, files, key, data_item, file_content)
        except RequestDataTooBig:
            self.errors[key] = RestDictError({'url': RestValidationError(
                ugettext('Response too large, maximum size is {} bytes').format(
                    pyston_settings.FILE_SIZE_LIMIT
                ))
            })
        except (RequestException, InvalidResponseStatusCode):
            self.errors[key] = RestDictError({'url': RestValidationError(
                ugettext('File is unreachable on the URL address')
            )})
        try:
            url_validator(url)
        except ValidationError as e:
            self.errors[key] = RestDictError({'url': RestValidationError(e.messages[0])})

    def _process_field(self, data, files, key, data_item):
        field = self.form.fields.get(key)
        if field and isinstance(field, FileField) and isinstance(data_item, dict):
            REQUIRED_ITEMS = {'content'}
            REQUIRED_URL_ITEMS = {'url'}

            if REQUIRED_ITEMS.issubset(set(data_item.keys())):
                for item in REQUIRED_ITEMS:
                    self._validate_not_empty(data_item, key, item)

                if not self.errors:
                    self._process_file_data_field(data, files, key, data_item)
            elif REQUIRED_URL_ITEMS.issubset(set(data_item.keys())):
                for item in REQUIRED_URL_ITEMS:
                    self._validate_not_empty(data_item, key, item)

                if not self.errors:
                    self._process_file_data_url_field(data, files, key, data_item)
            else:
                self.errors[key] = RestValidationError(
                    ugettext('File data item must contains {} or {}').format(
                        ', '.join(REQUIRED_ITEMS), ', '.join(REQUIRED_URL_ITEMS)
                    )
                )


class ModelResourceDataProcessor(DataProcessor):

    def __init__(self, resource, form, inst, via, partial_update):
        super(ModelResourceDataProcessor, self).__init__(resource, form)
        self.model = resource.model
        self.inst = inst
        self.via = resource._get_via(via)
        self.partial_update = partial_update


class MultipleDataProcessorMixin:

    INVALID_COLLECTION_EXCEPTION = {'error': _('Data must be a collection')}

    def _append_errors(self, key, operation, errors):
        self.errors[key] = self.errors.get(key, {})
        self.errors[key].update({operation: errors})


@data_preprocessors.register(BaseModelResource)
class ModelDataPreprocessor(ModelResourceDataProcessor):

    def _process_field(self, data, files, key, data_item):
        rest_field = getattr(self.form, 'related_fields', {}).get(key)

        if rest_field and not rest_field.is_reverse:
            try:
                data[key] = self.form.data[key] = rest_field.create_update_or_remove(
                    self.inst, data_item, self.via, self.request, self.partial_update, self.form
                )
            except RestError as ex:
                self.errors[key] = ex


@data_postprocessors.register(DjangoResource)
class ReverseDataPostprocessor(ModelResourceDataProcessor):

    def _process_field(self, data, files, key, data_item):
        rest_field = getattr(self.form, 'related_fields', {}).get(key)

        if rest_field and rest_field.is_reverse:
            try:
                data[key] = self.form.cleaned_data[key] = rest_field.create_update_or_remove(
                    self.inst, data_item, self.via, self.request, self.partial_update, self.form
                )
            except RestError as ex:
                self.errors[key] = ex



from pyston.filters.filters import OPERATORS
from pyston.contrib.dynamo.resource import BaseDynamoResource
from pyston.contrib.dynamo.filter import BaseDynamoFilter

from .models import Comment


class UserIDFilter(BaseDynamoFilter):

    allowed_operators = (OPERATORS.EQ,)

    def get_filter_term(self, value, operator_slug, request):
        return {'user_id__startswith': value}


class CommentDynamoResource(BaseDynamoResource):

    model = Comment
    fields = (
        'issue_id', 'user_id', 'content', 'is_public', 'priority'
    )
    filters = {
        'user_id': UserIDFilter,
    }

    can_read_obj = True
    range_key = 'user_id'

    def _get_hash_key(self):
        return self.kwargs.get('issue_pk')



from django.utils.safestring import mark_safe
from django.utils.translation import ugettext

from pyston.exception import RestException
from pyston.utils import rfs, LOOKUP_SEP
from pyston.utils.helpers import get_field_or_none, get_method_or_none
from pyston.serializer import get_resource_or_none

from .exceptions import OrderIdentifierError
from .parsers import DefaultOrderParser, OrderParserError
from .django_sorters import DjangoSorter


def get_allowed_order_fields_rfs_from_model(model):
    return rfs(model._rest_meta.extra_order_fields).join(rfs(model._rest_meta.order_fields))


class BaseModelOrderManager:
    """
    Order manager is used inside object resource for order response queryset according to input values.
    This is abstract class that provides methods to obtain concrete order strings from resource and model methods
    and fields.
    """

    def _get_real_field_name(self, resource, field_name):
        return resource.renamed_fields.get(field_name, field_name) if resource else field_name

    def _get_sorter_from_method(self, method, identifiers_prefix, identifiers, direction, model, resource, request,
                                order_fields_rfs):
        """
        :param method: method from which we can get order string.
        :param identifiers_prefix: because order strings are recursive if model relations property contains list of
               identifiers that was used for recursive searching the order string.
        :param identifiers: list of identifiers that conclusively identifies order string.
        :param direction: direction of ordering ASC or DESC.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param order_fields_rfs: RFS of fields that is allowed to order.
        :return: db order method string that is obtained from method.
        """
        if hasattr(method, 'order_by'):
            # If method has order_by attribute order string is being searched according to this value.
            order_identifiers = method.order_by.split(LOOKUP_SEP)
            # Because method must be inside allowed order fields RFS, we must add value order_by of the method
            # to the next RFS.
            next_order_fields_rfs = rfs(order_identifiers)
            return self._get_sorter_recursive(
                identifiers_prefix, order_identifiers, direction, model, resource, request, next_order_fields_rfs
            )
        if not identifiers_prefix and hasattr(method, 'sorter'):
            return method.sorter(identifiers_prefix + identifiers, direction)
        raise OrderIdentifierError

    def _get_sorter_from_resource(self, identifiers_prefix, identifiers, direction, model, resource, request,
                                  order_fields_rfs):
        """
        :param identifiers_prefix: because order strings are recursive if model relations property contains list of
               identifiers that was used for recursive searching the order string.
        :param identifiers: list of identifiers that conclusively identifies order string.
        :param direction: direction of ordering ASC or DESC.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param order_fields_rfs: RFS of fields that is allowed to order.
        :return: db order method string that is obtained from resource object.
        """
        full_identifiers_string = self._get_real_field_name(resource, LOOKUP_SEP.join(identifiers))

        resource_method = resource.get_method_returning_field_value(full_identifiers_string) if resource else None
        if full_identifiers_string in order_fields_rfs and resource_method:
            return self._get_sorter_from_method(resource_method, identifiers_prefix, identifiers, direction, model,
                                                resource, request, order_fields_rfs)

    def _get_sorter_from_model(self, identifiers_prefix, identifiers, direction, model, resource, request,
                               order_fields_rfs):
        return None

    def _get_order_fields_rfs(self, model, resource):
        return resource.get_order_fields_rfs() if resource else rfs()

    def _get_sorter_recursive(self, identifiers_prefix, identifiers, direction, model, resource, request,
                              extra_order_fields_rfs=None):
        """
        :param identifiers_prefix: because order strings are recursive if model relations property contains list of
               identifiers that was used for recursive searching the order string.
        :param identifiers: list of identifiers that conclusively identifies order string.
        :param direction: direction of ordering ASC or DESC.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param extra_order_fields_rfs: RFS of fields that is allowed to order.
        :return: method search resursice order string with order_string_from_model or order_string_from_resource
                 getters.
        """
        extra_order_fields_rfs = rfs() if extra_order_fields_rfs is None else extra_order_fields_rfs
        order_fields_rfs = extra_order_fields_rfs.join(self._get_order_fields_rfs(model, resource))
        order_string = (
            self._get_sorter_from_resource(
                identifiers_prefix, identifiers, direction, model, resource, request, order_fields_rfs) or
            self._get_sorter_from_model(
                identifiers_prefix, identifiers, direction, model, resource, request, order_fields_rfs)
        )
        if not order_string:
            raise OrderIdentifierError
        return order_string

    def get_sorter(self, identifiers, direction, resource, request):
        """
        :param identifiers: list of identifiers that conclusively identifies a order string.
        :param direction: direction of ordering ASC or DESC.
        :param resource: resource object.
        :param request: django HTTP request.
        :return: method returns filter string according to input identifiers, resource and request.
        """
        return self._get_sorter_recursive([], identifiers, direction, resource.model, resource, request)


class BaseDjangoOrderManager(BaseModelOrderManager):
    """
    Order manager is used inside model resource for order response queryset according to input values.
    This is abstract class that provides methods to obtain concrete order strings from resource and model methods
    and fields.
    """

    def _get_order_fields_rfs(self, model, resource):
        return resource.get_order_fields_rfs() if resource else get_allowed_order_fields_rfs_from_model(model)

    def _get_sorter_from_model(self, identifiers_prefix, identifiers, direction, model, resource, request,
                               order_fields_rfs):
        """
        :param identifiers_prefix: because order strings are recursive if model relations property contains list of
               identifiers that was used for recursive searching the order string.
        :param identifiers: list of identifiers that conclusively identifies order string.
        :param direction: direction of ordering ASC or DESC.
        :param model: django model class.
        :param resource: resource object.
        :param request: django HTTP request.
        :param order_fields_rfs: RFS of fields that is allowed to order.
        :return: db order method string that is obtained from model fields or methods.
        """
        current_identifier = self._get_real_field_name(resource, identifiers[0])
        identifiers_suffix = identifiers[1:]

        if current_identifier not in order_fields_rfs:
            raise OrderIdentifierError

        model_field = get_field_or_none(model, current_identifier)
        model_method = get_method_or_none(model, current_identifier)

        if model_field and not identifiers_suffix and (not model_field.is_relation or model_field.related_model):
            return DjangoSorter(identifiers_prefix + identifiers, direction)
        elif model_field and model_field.is_relation and model_field.related_model:
            next_model = model_field.related_model
            next_resource = get_resource_or_none(request, next_model, getattr(resource, 'resource_typemapper', None))
            return self._get_sorter_recursive(
                identifiers_prefix + [identifiers[0]], identifiers[1:], direction,
                next_model, next_resource, request, order_fields_rfs[current_identifier].subfieldset
            )
        elif model_method and not identifiers_suffix:
            return self._get_sorter_from_method(
                model_method, identifiers_prefix, identifiers, direction, model, resource, request, order_fields_rfs
            )


class BaseParserModelOrderManager(BaseModelOrderManager):

    parsers = [DefaultOrderParser()]

    def _get_sorters(self, parsed_order_terms, resource, request):
        """
        Converts order terms to sorter classes
        """
        sorters = []
        for ordering_term in parsed_order_terms:
            try:
                sorters.append(self.get_sorter(ordering_term.identifiers, ordering_term.direction, resource, request))
            except OrderIdentifierError:
                raise RestException(
                    mark_safe(ugettext('Invalid identifier of ordering "{}"').format(ordering_term.source))
                )
        return sorters

    def _convert_order_terms(self, sorters):
        """
        Converts sorters to the django query order strings.
        """
        return [sorter.get_order_term() for sorter in sorters]

    def _update_queryset(self, qs, sorters):
        """
        Update queryset for extra sorter class (it is used for annotations before ordering)
        """
        for sorter in sorters:
            qs = sorter.update_queryset(qs)
        return qs

    def _sort_queryset(self, qs, terms):
        raise NotImplementedError

    def _sort_with_parser(self, parser, resource, qs, request):
        try:
            parsed_order_terms = parser.parse(request)
            sorters = self._get_sorters(parsed_order_terms or (), resource, request)
            qs = self._update_queryset(qs, sorters)
            return self._sort_queryset(qs, self._convert_order_terms(sorters)) if sorters else qs
        except OrderParserError as ex:
            raise RestException(ex)

    def sort(self, resource, qs, request):
        for parser in self.parsers:
            qs = self._sort_with_parser(parser, resource, qs, request)
        return qs


class DjangoOrderManager(BaseParserModelOrderManager, BaseDjangoOrderManager):
    """
    Manager that uses parser to parse input order data to the list of order strings.
    """

    def _sort_queryset(self, qs, terms):
        return qs.order_by(*terms)



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-06-29 15:02
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('app', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='manual_created_date',
            field=models.DateTimeField(blank=True, null=True, verbose_name='manual created date'),
        ),
    ]



import json
import xml.dom.minidom

from decimal import Decimal
from uuid import uuid4
from datetime import date, datetime, time, timedelta

from collections import OrderedDict

from germanium.tools import assert_true, assert_equal

from unittest.case import TestCase

from app.models import User

from pyston.serializer import serialize


class DirectSerializationTestCase(TestCase):

    def test_serialization(self):
        for i in range(10):
            User.objects.create(is_superuser=True, email='test{}@test.cz'.format(i))
        assert_true(isinstance(json.loads((serialize(User.objects.all()))), list))
        assert_true(isinstance(json.loads((serialize(User.objects.first()))), dict))

        assert_equal(
            set(json.loads((serialize(User.objects.first()))).keys()),
            {'id', 'created_at', 'email', 'contract', 'solving_issue', 'first_name', 'last_name', 'manual_created_date'}
        )
        assert_equal(
            set(json.loads((serialize(User.objects.first(), ('id',)))).keys()),
            {'id'}
        )

        xml.dom.minidom.parseString(serialize(User.objects.first(), converter_name='xml'))

    def test_direct_serialization_to_csv_should_create_columns_according_to_required_fieldset(self):
        data = OrderedDict((('a', 1), ('b', 2)))
        assert_equal(
            serialize(data, converter_name='csv'),
            '\ufeff\r\n'
        )
        assert_equal(
            serialize(data, converter_name='csv', requested_fieldset=('a',)),
            '\ufeff"A"\r\n"1"\r\n'
        )
        assert_equal(
            serialize(data, converter_name='csv', requested_fieldset=('a', 'b')),
            '\ufeff"A";"B"\r\n"1";"2"\r\n'
        )
        assert_equal(
            serialize(data, converter_name='csv', requested_fieldset=('a', 'b', 'c')),
            '\ufeff"A";"B";"C"\r\n"1";"2";""\r\n'
        )

    def test_direct_serialization_to_python_format_should_return_only_base_python_data(self):
        now_value = datetime.now().replace(microsecond=0)
        today_value = date.today()
        timedelta_value = timedelta(days=5, hours=2, seconds=5)
        localtime_value = now_value.time()
        uuid_value = uuid4()
        decimal_value = Decimal('105.689')

        data = OrderedDict(
            now=now_value,
            today=today_value,
            timedelta=timedelta_value,
            localtime=localtime_value,
            uuid=uuid_value,
            decimal=decimal_value,
            set={1, 2, 3}
        )

        assert_equal(
            serialize(data, converter_name='python'),
            {
                'now': now_value.isoformat(),
                'today': today_value.isoformat(),
                'timedelta': '5 days, 2:00:05',
                'localtime': localtime_value.isoformat(),
                'uuid': str(uuid_value),
                'decimal': str(decimal_value),
                'set': [1, 2, 3]
            }
        )


from django.core.exceptions import FieldError, FieldDoesNotExist


def get_field_or_none(model, field_name):
    try:
        return model._meta.get_field(field_name)
    except FieldDoesNotExist:
        return None


def get_all_related_objects_from_model(model):
    return [
        f for f in model._meta.get_fields()
        if (f.one_to_many or f.one_to_one) and f.auto_created and not f.concrete
    ]


def get_concrete_field(model, field_name):
    field = {
        f.name: f for f in model._meta.get_fields()
        if f.concrete and (not f.is_relation or f.one_to_one or (f.many_to_one and f.related_model))
    }.get(field_name)
    if field:
        return field
    else:
        raise FieldDoesNotExist


def is_reverse_many_to_one(model, field_name):
    field = get_field_or_none(model, field_name)
    return field and field.auto_created and field.one_to_many


def is_reverse_one_to_one(model, field_name):
    field = get_field_or_none(model, field_name)
    return field and field.auto_created and field.one_to_one


def is_reverse_many_to_many(model, field_name):
    field = get_field_or_none(model, field_name)
    return field and field.auto_created and field.many_to_many


def is_many_to_one(model, field_name):
    field = get_field_or_none(model, field_name)
    return field and not field.auto_created and field.many_to_one


def is_one_to_one(model, field_name):
    field = get_field_or_none(model, field_name)
    return field and not field.auto_created and field.one_to_one


def is_many_to_many(model, field_name):
    field = get_field_or_none(model, field_name)
    return field and not field.auto_created and field.many_to_many


def is_relation(model, field_name):
    return (
        is_one_to_one(model, field_name) or
        is_many_to_one(model, field_name) or
        is_many_to_many(model, field_name) or
        is_reverse_one_to_one(model, field_name) or
        is_reverse_many_to_one(model, field_name) or
        is_reverse_many_to_many(model, field_name)
    )


def get_model_from_relation(model, field_name):
    try:
        related_model = model._meta.get_field(field_name).related_model
        if related_model:
            return related_model
        else:
            raise FieldError('field {} is not relation'.format(field_name))
    except FieldDoesNotExist:
        raise FieldError('field {} is not relation'.format(field_name))


def get_model_from_relation_or_none(model, field_name):
    try:
        return get_model_from_relation(model, field_name)
    except FieldError:
        return None


def get_reverse_field_name(model, field_name):
    """
    Gets reverse field name, but for reverse fields must be set related_name,
    therefore must be check if it was set by getting reverse field from the reverse model
    """
    try:
        model_field = model._meta.get_field(field_name)
        reverse_field_name = model_field.remote_field.name
        model_field.related_model._meta.get_field(reverse_field_name)
        return reverse_field_name
    except (FieldDoesNotExist, AttributeError):
        raise FieldError('field {} is not relation'.format(field_name))


def get_last_parent_pk_field_name(obj):
    for field in obj._meta.fields:
        if field.primary_key and (not field.is_relation or not field.auto_created):
            return field.name
    raise RuntimeError('Last parent field name was not found (cannot happen)')


def delete_cached_value(instance, field_name):
    """
    Backward compatibility between django 1 and 2. Django 2 has helper for removing cached values via field.
    """
    field = instance._meta.get_field(field_name)
    if hasattr(field, 'is_cached'):
        if field.is_cached(instance):
            field.delete_cached_value(instance)
    else:
        setattr(instance, field.get_cache_name(), None)



from django.template.defaultfilters import capfirst
from django.forms.utils import pretty_name

from pyston.utils import split_fields, is_match, LOOKUP_SEP, rfs
from pyston.utils.compatibility import get_model_from_relation_or_none


class Field:

    def __init__(self, key_path, label):
        self.key_path = key_path
        self.label = label

    def __str__(self):
        return capfirst(
            self.label if self.label is not None
            else pretty_name(' - '.join([key.replace('_', ' ').strip() for key in self.key_path]))
        )

    def __hash__(self):
        return hash(LOOKUP_SEP.join(self.key_path))

    def __eq__(self, other):
        return self.__str__() == other.__str__()

    def __ne__(self, other):
        return not self.__eq__(other)


class FieldsetGenerator:

    def __init__(self, resource=None, fields_string=None, direct_serialization=False):
        self.resource = resource
        self.fields_string = fields_string
        self.direct_serialization = direct_serialization

    def _get_resource(self, obj):
        from pyston.serializer import get_resource_or_none

        if self.resource:
            return get_resource_or_none(self.resource.request, obj, self.resource.resource_typemapper)
        else:
            return None

    def _get_allowed_fieldset(self):
        from pyston.resource import ModelResourceMixin

        # For security reasons only resource which defines allowed fields can be fully converted to the CSV/XLSX
        # or similar formats
        return self.resource.get_allowed_fields_rfs() if isinstance(self.resource, ModelResourceMixin) else rfs()

    def _parse_fields_string(self, fields_string):
        fields_string = fields_string or ''

        parsed_fields = []
        for field in split_fields(fields_string):
            if LOOKUP_SEP in field:
                field_name, subfields_string = field.split(LOOKUP_SEP, 1)
            elif is_match(r'^[^\(\)]+\(.+\)$', field):
                field_name, subfields_string = field[:len(field) - 1].split('(', 1)
            else:
                field_name, subfields_string = field, None

            parsed_fields.append((field_name, subfields_string))
        return parsed_fields

    def _recursive_generator(self, fields, fields_string, model=None, key_path=None, extended_fieldset=None):
        key_path = key_path or []

        allowed_fieldset = self._get_allowed_fieldset()
        if extended_fieldset:
            allowed_fieldset.join(extended_fieldset)

        parsed_fields = [
            (field_name, subfields_string) for field_name, subfields_string in self._parse_fields_string(fields_string)
            if field_name in allowed_fieldset or self.direct_serialization
        ]

        for field_name, subfields_string in parsed_fields:
            self._recursive_generator(
                fields, subfields_string, get_model_from_relation_or_none(model, field_name) if model else None,
                key_path + [field_name],
                extended_fieldset=allowed_fieldset[field_name].subfieldset if allowed_fieldset[field_name] else None
            )
        if not parsed_fields and key_path:
            fields.append(
                Field(key_path, self.resource.get_field_label(LOOKUP_SEP.join(key_path)) if self.resource else None)
            )

    def generate(self):
        fields = []
        self._recursive_generator(fields, self.fields_string, getattr(self.resource, 'model', None))
        return fields



import re

from collections import OrderedDict

from django.template.defaultfilters import lower
from django.db import models
from django.utils.encoding import force_text

from copy import deepcopy


LOOKUP_SEP = '__'


def coerce_rest_request_method(request):
    """
    Django doesn't particularly understand REST.
    In case we send data over PUT, Django won't
    actually look at the data and load it. We need
    to twist its arm here.

    The try/except abominiation here is due to a bug
    in mod_python. This should fix it.
    """
    if request.method in {'PUT', 'PATCH'}:
        # Bug fix: if _load_post_and_files has already been called, for
        # example by middleware accessing request.POST, the below code to
        # pretend the request is a POST instead of a PUT will be too late
        # to make a difference. Also calling _load_post_and_files will result
        # in the following exception:
        #   AttributeError: You cannot set the upload handlers after the upload has been processed.
        # The fix is to check for the presence of the _post field which is set
        # the first time _load_post_and_files is called (both by wsgi.py and
        # modpython.py). If it's set, the request has to be 'reset' to redo
        # the query value parsing in POST mode.
        if hasattr(request, '_post'):
            del request._post
            del request._files

        tmp_request_method = request.method
        try:
            request.method = 'POST'
            request._load_post_and_files()
            request.method = tmp_request_method
        except AttributeError:
            request.META['REQUEST_METHOD'] = 'POST'
            request._load_post_and_files()
            request.META['REQUEST_METHOD'] = tmp_request_method

        setattr(request, tmp_request_method, request.POST)


def model_all_available_fields(model):
    return {field.name for field in model._meta.fields}


def flat_list(list_obj):
    flat_list_obj = []
    for val in list_obj:
        if isinstance(val, (list, tuple)):
            flat_list_obj.append(val[0])
        else:
            flat_list_obj.append(val)
    return flat_list_obj


class JsonObj(dict):

    def __setattr__(self, name, value):
        self[name] = value


def model_resources_to_dict():
    from pyston.resource import resource_tracker

    model_resources = {}
    for resource in resource_tracker:
        if hasattr(resource, 'model') and issubclass(resource.model, models.Model):
            model = resource.model
            model_label = lower('{}.{}'.format(model._meta.app_label, model._meta.object_name))
            model_resources[model_label] = resource
    return model_resources


def set_rest_context_to_request(request, mapping):
    context = {}
    for key, (header_key, queryset_key) in mapping.items():
        val = request.GET.get(queryset_key, request.META.get(header_key))
        if val:
            context[key] = val
    request._rest_context = context


def is_match(regex, text):
    pattern = re.compile(regex)
    return pattern.search(text) is not None


def split_fields(fields_string):
    brackets = 0

    field = ''
    for char in fields_string:
        if char == ',' and not brackets:
            field = field.strip()
            if field:
                yield field
            field = ''
            continue

        if char == '(':
            brackets += 1

        if char == ')':
            brackets -= 1

        field += char

    field = field.strip()
    if field:
        yield field


class RestField:

    def __init__(self, name, subfieldset=None):
        assert isinstance(name, str)
        assert subfieldset is None or isinstance(subfieldset, RestFieldset)

        self.name = name
        self.subfieldset = subfieldset or RestFieldset()

    def __deepcopy__(self, memo):
        return self.__class__(self.name, deepcopy(self.subfieldset))

    def join(self, rest_field):
        self.subfieldset = self.subfieldset.join(rest_field.subfieldset)
        return self

    def intersection(self, rest_field):
        self.subfieldset = self.subfieldset.intersection(rest_field.subfieldset)
        return self

    def __str__(self):
        if self.subfieldset:
            return '{}({})'.format(self.name, self.subfieldset)
        return force_text(self.name)


class RestFieldset:

    @classmethod
    def create_from_string(cls, fields_string):
        fields = []
        for field in split_fields(fields_string):
            if is_match(r'^[^\(\)]+\(.+\)$', field):
                field_name, subfields_string = field[:len(field) - 1].split('(', 1)
                if LOOKUP_SEP in field_name:
                    field_name, subfields_string = field.split(LOOKUP_SEP, 1)

                subfieldset = RFS.create_from_string(subfields_string)
            else:
                field_name = field
                subfieldset = None
                if LOOKUP_SEP in field_name:
                    field_name, subfields_string = field.split(LOOKUP_SEP, 1)
                    subfieldset = RFS.create_from_string(subfields_string)

            fields.append(RestField(field_name, subfieldset))
        return RestFieldset(*fields)

    @classmethod
    def _create_field_from_list(cls, field):
        field_name, subfield_list = field

        return RestField(field_name, cls.create_from_list(subfield_list))

    @classmethod
    def _create_field_from_string(cls, field):
        if LOOKUP_SEP in field:
            field_name, field_child = field.split(LOOKUP_SEP, 1)
            return RestField(field_name, cls.create_from_list((field_child,)))
        else:
            return RestField(field)

    @classmethod
    def create_from_list(cls, fields_list=None):
        if isinstance(fields_list, RestFieldset):
            return deepcopy(fields_list)

        fields = []
        for field in fields_list or ():
            if isinstance(field, (list, tuple)):
                fields.append(cls._create_field_from_list(field))
            elif isinstance(field, str):
                fields.append(cls._create_field_from_string(field))
            else:
                raise ValueError('field can be only list, tuple or string ({} [{}])'.format(field, type(field)))

        return RestFieldset(*fields)

    def __init__(self, *fields):
        self.fields_map = OrderedDict()
        for field in fields:
            if not isinstance(field, RestField):
                field = RestField(field)
            self.append(field)

    @property
    def fields(self):
        return self.fields_map.values()

    def join(self, rest_fieldset):
        if isinstance(rest_fieldset, (list, tuple, set)):
            rest_fieldset = self.create_from_list(rest_fieldset)

        assert isinstance(rest_fieldset, RestFieldset)

        for rf in rest_fieldset.fields:
            if rf.name not in self.fields_map:
                self.fields_map[rf.name] = deepcopy(rf)
            else:
                self.fields_map[rf.name] = self.fields_map[rf.name].join(rf)

        return self

    def intersection(self, rest_fieldset):
        assert isinstance(rest_fieldset, RestFieldset)

        fields_map = self.fields_map
        self.fields_map = OrderedDict()

        for name, rf in fields_map.items():
            if name in rest_fieldset.fields_map:
                self.append(rf.intersection(rest_fieldset.fields_map[name]))

        return self

    def subtract(self, rest_fieldset):
        if isinstance(rest_fieldset, (list, tuple, set)):
            rest_fieldset = RFS(*rest_fieldset)

        assert isinstance(rest_fieldset, RestFieldset)

        fields_map = self.fields_map
        self.fields_map = OrderedDict()

        for name, rf in fields_map.items():
            if name not in rest_fieldset.fields_map:
                self.fields_map[name] = rf

        return self

    def __deepcopy__(self, memo):
        return self.__class__(*map(deepcopy, self.fields))

    def __str__(self):
        return ','.join(map(force_text, self.fields))

    def __add__(self, rest_fieldset):
        a_rfs = deepcopy(self)
        return a_rfs.join(rest_fieldset)

    def __bool__(self):
        return bool(self.fields_map)
    __nonzero__ = __bool__

    def get(self, key):
        return self.fields_map.get(key)

    def append(self, field):
        if isinstance(field, RestField):
            rest_field = field
        elif isinstance(field, str):
            rest_field = self._create_field_from_string(field)
        elif isinstance(field, (list, tuple)):
            rest_field = self._create_field_from_list(field)
        else:
            raise ValueError('field can be only list, tuple or string ({} [{}])'.format(field, type(field)))

        if rest_field.name in self.fields_map:
            rest_field = self.fields_map[rest_field.name].join(rest_field)

        self.fields_map[rest_field.name] = rest_field
        return self

    def update(self, rest_fieldset):
        if isinstance(rest_fieldset, (list, tuple, set)):
            rest_fieldset = self.create_from_list(rest_fieldset)

        assert isinstance(rest_fieldset, RestFieldset)

        for rf in rest_fieldset.fields:
            rf = deepcopy(rf)
            if rf.name not in self.fields_map:
                self.fields_map[rf.name] = rf
            else:
                self.fields_map[rf.name] = self.fields_map[rf.name].join(rf)

        return self

    def flat(self):
        return set(self.fields_map.keys())

    def __contains__(self, key):
        return key in self.fields_map

    def __getitem__(self, key):
        return self.get(key)


RF = RestField
RFS = RestFieldset
rfs = RFS.create_from_list



import re

from pyston.converters import JsonConverter, is_collection


def to_camel_case(snake_str):
    if snake_str.startswith('_'):
        return '_' + to_camel_case(snake_str[1:])
    else:
        components = snake_str.split('_')
        # We capitalize the first letter of each component except the first one
        # with the 'title' method and join them together.
        return components[0] + ''.join(x.title() for x in components[1:])


def to_snake_case(name):
    s1 = re.sub('(.)([A-Z])', r'\1_\2', name)
    return re.sub('([^_])([A-Z])', r'\1_\2', s1).lower()


class JsonCamelCaseConverter(JsonConverter):

    def _encode_snake_to_camel(self, data):
        from pyston.serializer import LAZY_SERIALIZERS

        if isinstance(data, LAZY_SERIALIZERS):
            return self._encode_snake_to_camel(data.serialize())
        elif is_collection(data):
            return (self._encode_snake_to_camel(item) for item in data)
        elif isinstance(data, dict):
            return {to_camel_case(key): self._encode_snake_to_camel(val) for key, val in data.items()}
        else:
            return data

    def _decode_camel_to_snake(self, data):
        if is_collection(data):
            return [self._decode_camel_to_snake(item) for item in data]
        elif isinstance(data, dict):
            return {to_snake_case(key): self._decode_camel_to_snake(val) for key, val in data.items()}
        else:
            return data

    def _encode_to_stream(self, output_stream, data, options=None, **kwargs):
        super()._encode_to_stream(output_stream, self._encode_snake_to_camel(data), options=None, **kwargs)

    def _decode(self, data, **kwargs):
        return self._decode_camel_to_snake(super(JsonCamelCaseConverter, self)._decode(data))



from pyston.utils import LOOKUP_SEP

from .utils import DIRECTION


class OrderParserError(Exception):
    """
    Exception that is raised if order input is invalid.
    """
    pass


class OrderTerm:
    """
    Simple order term that contains order identifiers list, direction and source inpout value which is used to assemble
    error messages.
    """

    def __init__(self, identifiers, direction, source):
        self.identifiers = identifiers
        self.direction = direction
        self.source = source


class OrderParser:
    """
    Abstract order parser.
    """

    def parse(self, request):
        """
        :param request: Django HTTP request.
        :return: returns list of order terms.
        """
        raise NotImplementedError


class DefaultOrderParser:
    """
    Default order parser that accepts filter.
    E.q.:
        /api/user?order=first_name,-created_at
    """

    def _clean_order_term(self, ordering_string):
        ordering_string = ordering_string.strip()
        if ordering_string.startswith('-'):
            direction = DIRECTION.DESC
            ordering_string = ordering_string[1:]
        else:
            direction = DIRECTION.ASC

        identifiers = ordering_string.split(LOOKUP_SEP)
        return OrderTerm(identifiers, direction, ordering_string)

    def parse(self, request):
        order_fields = request._rest_context.get('order')
        if order_fields:
            return (self._clean_order_term(ordering_string) for ordering_string in order_fields.split(','))
        else:
            return None



from pyston.order.managers import BaseParserModelOrderManager
from pyston.order.sorters import BaseSorter
from pyston.order.utils import DIRECTION
from pyston.utils import LOOKUP_SEP


class ElasticsearchSorter(BaseSorter):

    def get_order_term(self):
        term = LOOKUP_SEP.join(self.identifiers)
        return f'-{term}' if self.direction == DIRECTION.DESC else term


class ElasticsearchOrderManager(BaseParserModelOrderManager):

    order_by_field_name = {'boolean', 'date', 'keyword', 'float', 'integer'}

    def _sort_queryset(self, qs, terms):
        return qs.sort(*terms)

    def _get_sorter_from_model(self, identifiers_prefix, identifiers, direction, model, resource, request,
                               order_fields_rfs):
        if len(identifiers) == 1:
            current_identifier = self._get_real_field_name(resource, identifiers[0])
            try:
                field = model._doc_type.mapping.properties[current_identifier]
                field_name = field.name
                if hasattr(field, 'builtin_type') and field_name not in self.order_by_field_name:
                    field_name = field.builtin_type.name
                if field_name in self.order_by_field_name:
                    return ElasticsearchSorter(identifiers_prefix + identifiers, direction)
            except KeyError:
                pass
        return None
